import os
import string
from pathlib import Path

from .types import TimeAlignedType


def _get_duration(name: str) -> float:
    import soundfile

    try:
        return soundfile.info(name).duration
    except Exception as e:
        raise OSError(f"Error reading {name}: {e}") from e


def load_text(audio: str | os.PathLike[str]) -> TimeAlignedType | None:
    """Load time-aligned text data given a L2-ARCTIC audio file.

    :param audio: Path to the L2-ARCTIC audio file.
    :return: A TimeAlignedType object.
    """
    file = Path(audio).parent.parent / "transcript" / (Path(audio).stem + ".txt")
    if not os.path.exists(file):
        return None

    with open(file, encoding="utf-8") as f:
        line = f.read()

    return TimeAlignedType(
        0,
        _get_duration(str(audio)),
        line.strip().lower().translate(str.maketrans("", "", string.punctuation)),
    )


def load_words(audio: str | os.PathLike[str]) -> list[TimeAlignedType] | None:
    """Load time-aligned word data given a L2-ARCTIC audio file.

    :param audio: Path to the L2-ARCTIC audio file.
    :return: A list of TimeAlignedType objects.
    """
    return _load_ta(audio, "words")


def load_phonemes(audio: str | os.PathLike[str]) -> list[TimeAlignedType] | None:
    """Load time-aligned phonemes data given a L2-ARCTIC audio file.

    :param audio: Path to the L2-ARCTIC audio file.
    :return: A list of TimeAlignedType objects.
    """
    return _load_ta(audio, "phones")


def _load_ta(audio: str | os.PathLike[str], tier: str) -> list[TimeAlignedType] | None:
    from praatio import textgrid
    from praatio.utilities.constants import Interval

    file = Path(audio).parent.parent / "textgrid" / (Path(audio).stem + ".TextGrid")
    if not os.path.exists(file):
        return None

    tg = textgrid.openTextgrid(str(file), includeEmptyIntervals=False)
    if tier not in tg.tierNames:
        return None

    entries: list[TimeAlignedType] = []
    for entry in tg.getTier(tier).entries:
        if isinstance(entry, Interval):
            entries.append(TimeAlignedType(text=entry.label, start=entry.start, end=entry.end))

    return entries


def load_annotations(
    audio: str | os.PathLike[str],
) -> dict[str, list[TimeAlignedType]] | None:
    """Load time-aligned annotation data given a L2-ARCTIC audio file.

    :param audio: Path to the L2-ARCTIC audio file.
    :return: A dictionary of a list of TimeAlignedType objects.
    """
    from praatio import textgrid
    from praatio.utilities.constants import Interval

    file = Path(audio).parent.parent / "annotation" / (Path(audio).stem + ".TextGrid")
    if not os.path.exists(file):
        return None

    tg = textgrid.openTextgrid(str(file), includeEmptyIntervals=False)
    result: dict[str, list[TimeAlignedType]] = {}
    for tier in tg.tierNames:
        entries: list[TimeAlignedType] = []
        for entry in tg.getTier(tier).entries:
            if isinstance(entry, Interval):
                entries.append(TimeAlignedType(text=entry.label, start=entry.start, end=entry.end))
        result[tier] = entries

    return result


def load_speakers(input_dir: Path) -> dict:
    speakers = {}
    with open(input_dir / "readme-download.txt") as file:
        processing = False
        for line in file:
            if not processing and line.startswith("|---|"):
                processing = True
                continue

            if processing:
                if line.startswith("|**Total**|"):
                    break
                else:
                    fields = line.strip().split("|")
                    speaker_id = fields[1]
                    gender = fields[2]
                    dialect = fields[3]
                    speakers[speaker_id] = {"gender": gender, "dialect": dialect}

    return speakers



from ...datatypes import Truth
from ..mixdb import MixtureDatabase


def target_f_validate(_config: dict) -> None:
    pass


def target_f_parameters(feature: str, _num_classes: int, _config: dict) -> int:
    from pyaaware import ForwardTransform
    from pyaaware import feature_forward_transform_config

    ft = ForwardTransform(**feature_forward_transform_config(feature))

    if ft.ttype == "tdac-co":
        return ft.bins

    return ft.bins * 2


def target_f(mixdb: MixtureDatabase, m_id: int, category: str, _config: dict) -> Truth:
    """Frequency domain target truth function

    Calculates the true transform of the target using the STFT
    configuration defined by the feature. This will include a
    forward transform window if defined by the feature.

    Output shape: [:, 2 * bins] (target stacked real, imag) or
                  [:, bins] (target real only for tdac-co)
    """
    import torch
    from pyaaware import ForwardTransform
    from pyaaware import feature_forward_transform_config

    ft = ForwardTransform(**feature_forward_transform_config(mixdb.feature))

    target_audio = torch.from_numpy(mixdb.mixture_sources(m_id)[category])

    target_freq = ft.execute_all(target_audio)[0].numpy()
    return _stack_real_imag(target_freq, ft.ttype)


def target_mixture_f_validate(_config: dict) -> None:
    pass


def target_mixture_f_parameters(feature: str, _num_classes: int, _config: dict) -> int:
    from pyaaware import ForwardTransform
    from pyaaware import feature_forward_transform_config

    ft = ForwardTransform(**feature_forward_transform_config(feature))

    if ft.ttype == "tdac-co":
        return ft.bins * 2

    return ft.bins * 4


def target_mixture_f(mixdb: MixtureDatabase, m_id: int, category: str, _config: dict) -> Truth:
    """Frequency domain target and mixture truth function

    Calculates the true transform of the target and the mixture
    using the STFT configuration defined by the feature. This
    will include a forward transform window if defined by the
    feature.

    Output shape: [:, 4 * bins] (target stacked real, imag; mixture stacked real, imag) or
                  [:, 2 * bins] (target real; mixture real for tdac-co)
    """
    import numpy as np
    import torch
    from pyaaware import ForwardTransform
    from pyaaware import feature_forward_transform_config

    ft = ForwardTransform(**feature_forward_transform_config(mixdb.feature))

    target_audio = torch.from_numpy(mixdb.mixture_sources(m_id)[category])
    mixture_audio = torch.from_numpy(mixdb.mixture_mixture(m_id))

    target_freq = ft.execute_all(torch.from_numpy(target_audio))[0].numpy()
    mixture_freq = ft.execute_all(torch.from_numpy(mixture_audio))[0].numpy()

    frames, bins = target_freq.shape
    truth = np.empty((frames, bins * 4), dtype=np.float32)
    truth[:, : bins * 2] = _stack_real_imag(target_freq, ft.ttype)
    truth[:, bins * 2 :] = _stack_real_imag(mixture_freq, ft.ttype)
    return truth


def target_swin_f_validate(_config: dict) -> None:
    pass


def target_swin_f_parameters(feature: str, _num_classes: int, _config: dict) -> int:
    from pyaaware import ForwardTransform
    from pyaaware import feature_forward_transform_config

    return ForwardTransform(**feature_forward_transform_config(feature)).bins * 2


def target_swin_f(mixdb: MixtureDatabase, m_id: int, category: str, _config: dict) -> Truth:
    """Frequency domain target with synthesis window truth function

    Calculates the true transform of the target using the STFT
    configuration defined by the feature. This will include a
    forward transform window if defined by the feature and also
    the inverse transform (or synthesis) window.

    Output shape: [:, 2 * bins] (stacked real, imag)
    """
    import numpy as np
    import torch
    from pyaaware import ForwardTransform
    from pyaaware import InverseTransform
    from pyaaware import feature_forward_transform_config
    from pyaaware import feature_inverse_transform_config

    from ...utils.stacked_complex import stack_complex

    ft = ForwardTransform(**feature_forward_transform_config(mixdb.feature))
    it = InverseTransform(**feature_inverse_transform_config(mixdb.feature))

    target_audio = mixdb.mixture_sources(m_id)[category]

    truth = np.empty((len(target_audio) // ft.overlap, ft.bins * 2), dtype=np.float32)
    for idx, offset in enumerate(range(0, len(target_audio), ft.overlap)):
        audio_frame = torch.from_numpy(np.multiply(target_audio[offset : offset + ft.overlap], it.window))
        target_freq = ft.execute(audio_frame)[0].numpy()
        truth[idx] = stack_complex(target_freq)

    return truth


def _stack_real_imag(data: Truth, ttype: str) -> Truth:
    import numpy as np

    from ...utils.stacked_complex import stack_complex

    if ttype == "tdac-co":
        return np.real(data)

    return stack_complex(data)



def human_readable_size(num: float, decimal_places: int = 3, suffix: str = "B") -> str:
    """Convert number into string with units"""
    for unit in ("", "k", "M", "G", "T", "P", "E", "Z"):
        if abs(num) < 1024.0:
            return f"{num:.{decimal_places}f} {unit}{suffix}"
        num /= 1024.0
    return f"{num:.{decimal_places}f} Y{suffix}"



from pathlib import Path

from praatio import textgrid
from praatio.utilities.constants import Interval

from .types import TimeAlignedType


def create_textgrid(
    prompt: Path,
    output_dir: Path,
    text: TimeAlignedType | None = None,
    words: list[TimeAlignedType] | None = None,
    phonemes: list[TimeAlignedType] | None = None,
) -> None:
    if text is None and words is None and phonemes is None:
        return

    min_t, max_t = _get_min_max({"phonemes": phonemes, "text": text, "words": words})

    tg = textgrid.Textgrid()

    if text is not None:
        entries = [Interval(text.start, text.end, text.text)]
        text_tier = textgrid.IntervalTier("text", entries, min_t, max_t)
        tg.addTier(text_tier)

    if words is not None:
        entries = []
        for word in words:
            entries.append(Interval(word.start, word.end, word.text))
        words_tier = textgrid.IntervalTier("words", entries, min_t, max_t)
        tg.addTier(words_tier)

    if phonemes is not None:
        entries = []
        for phoneme in phonemes:
            entries.append(Interval(phoneme.start, phoneme.end, phoneme.text))
        phonemes_tier = textgrid.IntervalTier("phonemes", entries, min_t, max_t)
        tg.addTier(phonemes_tier)

    output_filename = str(output_dir / prompt.stem) + ".TextGrid"
    tg.save(output_filename, format="long_textgrid", includeBlankSpaces=True)


def _get_min_max(tiers: dict[str, TimeAlignedType | list[TimeAlignedType] | None]) -> tuple[float, float]:
    starts = []
    ends = []
    for tier in tiers.values():
        if tier is None:
            continue
        if isinstance(tier, TimeAlignedType):
            starts.append(tier.start)
            ends.append(tier.end)
        else:
            starts.append(tier[0].start)
            ends.append(tier[-1].end)

    return min(starts), max(ends)


def annotate_textgrid(
    tiers: dict[str, TimeAlignedType | list[TimeAlignedType] | None] | None, prompt: Path, output_dir: Path
) -> None:
    import os

    if tiers is None:
        return

    file = Path(output_dir / prompt.stem).with_suffix(".TextGrid")
    if not os.path.exists(file):
        tg = textgrid.Textgrid()
        min_t, max_t = _get_min_max(tiers)
    else:
        tg = textgrid.openTextgrid(str(file), includeEmptyIntervals=False)
        min_t = tg.minTimestamp
        max_t = tg.maxTimestamp

    for k, v in tiers.items():
        if v is None:
            continue
        entries = [Interval(entry.start, entry.end, entry.text) for entry in v]
        if k == "phones":
            name = "annotation_phonemes"
        else:
            name = "annotation_" + k
        tg.addTier(textgrid.IntervalTier(name, entries, min_t, max_t))

    tg.save(str(file), format="long_textgrid", includeBlankSpaces=True)



"""
Parse config rules

sai_expand(0, 5)
sai_expand(.0, 5)
sai_expand(0., 5)
sai_expand(0.0, 5)
sai_expand(0, .5)
sai_expand(.0, .5)
sai_expand(0., .5)
sai_expand(0.0, .5)
sai_expand(0, 5.)
sai_expand(.0, 5.)
sai_expand(0., 5.)
sai_expand(0.0, 5.)
sai_expand(0, 5.0)
sai_expand(.0, 5.0)
sai_expand(0., 5.0)
sai_expand(0.0, 5.0)

sai_rand(-1, 1)
sai_choose(repeat=False)
sai_choose(repeat=False, tag=tag)
sai_sequence()
sai_sequence(tag=tag)

sai_expand(0, sai_rand(-1, 1))
sai_expand(sai_rand(0, 4), 0)
sai_expand(sai_rand(0, 4), sai_rand(-1, 1))

sai_choose(num=1, unique=speaker_id, repeat=False)
sai_choose(num=1, repeat=False)
sai_choose(num=1, unique=speaker_id, repeat=True)
sai_choose(num=1, repeat=True)
sai_sequence(num=0)
sai_sequence(num=0, unique=speaker_id)

"""

from dataclasses import dataclass


@dataclass
class Match:
    group: str
    span: tuple[int, int]

    def start(self) -> int:
        return self.span[0]

    def end(self) -> int:
        return self.span[1]


def find_sai_expand(text: str) -> list[Match]:
    import re

    results = []
    matches = re.finditer(r"sai_expand\(", text)
    if matches:
        for match in matches:
            s = match.start()
            e = match.end()
            num_lparen = 1
            while num_lparen != 0 and e < len(text):
                if text[e] == "(":
                    num_lparen += 1
                elif text[e] == ")":
                    num_lparen -= 1
                e += 1
            if num_lparen != 0:
                raise ValueError(f"Unbalanced parenthesis in '{text}'")

            results.append(Match(group=text[s:e], span=(s, e)))

    return results


def parse_sai_expand(text: str) -> list[str]:
    """
    expand_syntax ::= expand_keyword lparen expand_item next_expand_item+ rparen
    expand_keyword ::= sai_expand
    lparen ::= (
    expand_item ::= expand_syntax | rand_syntax | number
    rand_syntax ::= rand_keyword lparen rand_item comma rand_item rparen
    rand_item ::= real number
    comma ::= ,
    number ::= real number or signed integer
    next_item ::= comma item
    rparen ::= )
    """
    import pyparsing as pp

    lparen = pp.Literal("(")
    rparen = pp.Literal(")")
    comma = pp.Literal(",")

    real_number = pp.pyparsing_common.real
    signed_integer = pp.pyparsing_common.signed_integer
    number = real_number | signed_integer

    identifier = pp.Word(pp.alphanums + "_.-")

    rand_literal = pp.Literal("sai_rand")
    rand_expression = (rand_literal + lparen + number + comma + number + rparen).set_parse_action(
        lambda tokens: "".join(map(str, tokens))
    )

    expand_literal = pp.Literal("sai_expand")
    expand_args = pp.DelimitedList(rand_expression | identifier, min=1)
    expand_expression = expand_literal + lparen + expand_args("args") + rparen

    try:
        result = expand_expression.parse_string(text)
    except pp.ParseException as e:
        raise ValueError(f"Could not parse '{text}'") from e

    return list(result.args)


def sai_expand(text: str) -> list[str]:
    # initialize with input
    expanded = [text]

    # look for pattern
    matches = find_sai_expand(text)

    # if not found, early exit
    if not matches:
        return expanded

    # remove entry we are expanding
    expanded.pop()

    # start with the innermost match
    match = matches[-1]
    prelude = text[: match.start()]
    postlude = text[match.end() :]

    # loop over parsed expand values
    for value in parse_sai_expand(match.group):
        # extend result with expand of replacement (for handling multiple expands in a single rule)
        expanded.extend(sai_expand(prelude + value + postlude))

    return expanded



import numpy as np

from ..datatypes import AudioF
from ..datatypes import Segsnr
from ..datatypes import SnrFBinMetrics
from ..datatypes import SnrFMetrics


def calc_segsnr_f(segsnr_f: Segsnr) -> SnrFMetrics:
    """Calculate metrics of snr_f truth data.

    Includes mean and standard deviation of the linear values (usually energy)
    and mean and standard deviation of the dB values (10 * log10).
    """
    if np.count_nonzero(segsnr_f) == 0:
        # If all entries are zeros
        return SnrFMetrics(0, 0, -np.inf, 0)

    tmp = np.ma.array(segsnr_f, mask=np.logical_not(np.isfinite(segsnr_f)))
    if np.ma.count_masked(tmp) == np.ma.size(tmp, axis=0):
        # If all entries are infinite
        return SnrFMetrics(np.inf, 0, np.inf, 0)

    snr_mean = np.mean(tmp, axis=0)
    snr_std = np.std(tmp, axis=0)

    tmp = 10 * np.ma.log10(tmp)
    if np.ma.count_masked(tmp) == np.ma.size(tmp, axis=0):
        # If all entries are masked, special case where all inputs are either 0 or infinite
        snr_db_mean = -np.inf
        snr_db_std = np.inf
    else:
        snr_db_mean = np.mean(tmp, axis=0)
        snr_db_std = np.std(tmp, axis=0)

    return SnrFMetrics(snr_mean, snr_std, snr_db_mean, snr_db_std)


def calc_segsnr_f_bin(target_f: AudioF, noise_f: AudioF) -> SnrFBinMetrics:
    """Calculate per-bin segmental SNR metrics.

    Includes per-bin mean and standard deviation of the linear values
    and mean and standard deviation of the dB values.
    """
    if target_f.ndim != 2 and noise_f.ndim != 2:
        raise ValueError("target_f and noise_f must have 2 dimensions")

    segsnr_f = (np.abs(target_f) ** 2) / (np.abs(noise_f) ** 2 + np.finfo(np.float32).eps)

    frames, bins = segsnr_f.shape
    if np.count_nonzero(segsnr_f) == 0:
        # If all entries are zeros
        return SnrFBinMetrics(np.zeros(bins), np.zeros(bins), -np.inf * np.ones(bins), np.zeros(bins))

    tmp = np.ma.array(segsnr_f, mask=np.logical_not(np.isfinite(segsnr_f)))
    if np.ma.count_masked(tmp) == np.ma.size(tmp, axis=0):
        # If all entries are infinite
        return SnrFBinMetrics(
            np.inf * np.ones(bins),
            np.zeros(bins),
            np.inf * np.ones(bins),
            np.zeros(bins),
        )

    snr_mean = np.mean(tmp, axis=0)
    snr_std = np.std(tmp, axis=0)

    tmp = 10 * np.ma.log10(tmp)
    if np.ma.count_masked(tmp) == np.ma.size(tmp, axis=0):
        # If all entries are masked, special case where all inputs are either 0 or infinite
        snr_db_mean = -np.inf * np.ones(bins)
        snr_db_std = np.inf * np.ones(bins)
    else:
        snr_db_mean = np.mean(tmp, axis=0)
        snr_db_std = np.std(tmp, axis=0)

    return SnrFBinMetrics(
        np.ma.getdata(snr_mean),
        np.ma.getdata(snr_std),
        np.ma.getdata(snr_db_mean),
        np.ma.getdata(snr_db_std),
    )



"""sonusai genmetrics

usage: genmetrics [-hvusd] [-i MIXID] [-n INCLUDE] [-x EXCLUDE] [-p NUMPROC] LOC

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -i MIXID, --mixid MIXID         Mixture ID(s) to generate. [default: *].
    -n INCLUDE, --include INCLUDE   Metrics to include. [default: all]
    -x EXCLUDE, --exclude EXCLUDE   Metrics to exclude. [default: none]
    -p NUMPROC, --nproc NUMPROC     Number of parallel processes to use. Default single thread.
    -u, --update                    Update metrics (do not regenerate existing metrics).
    -s, --supported                 Show list of supported metrics.
    -d, --dryrun                    Show list of metrics that will be generated and exit.

Calculate speech enhancement metrics of SonusAI mixture data in LOC.

Inputs:
    LOC         A SonusAI mixture database directory.
    MIXID       A glob of mixture ID(s) to generate.
    INCLUDE     Comma separated list of metrics to include. Can be "all" or
                any of the supported metrics or glob(s).
    EXCLUDE     Comma separated list of metrics to exclude. Can be "none" or
                any of the supported metrics or glob(s)

Note: The default include of "all" excludes the generation of ASR metrics,
i.e., "*asr*,*wer*". However, if include is manually specified to something other than "all",
then this behavior is overridden.

Similarly, the default exclude of "none" excludes the generation of ASR metrics,
i.e., "*asr*,*wer*". However, if exclude is manually specified to something other than "none",
then this behavior is also overridden.

Examples:

Generate all available mxwer metrics (as determined by mixdb asr_configs parameter):
> sonusai genmetrics -n"mxwer*" mixdb_loc

Generate only mxwer.faster metrics:
> sonusai genmetrics -n"mxwer.faster" mixdb_loc

Generate only faster metrics:
> sonusai genmetrics -n"*faster" mixdb_loc

Generate all available metrics except for mxcovl
> sonusai genmetrics -x"mxcovl" mixdb_loc

"""


def _process_mixture(mixid: int, location: str, metrics: list[str], update: bool = False) -> set[str]:
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import write_cached_data
    from sonusai.mixture import write_mixture_metadata

    mixdb = MixtureDatabase(location)
    results = mixdb.mixture_metrics(m_id=mixid, metrics=metrics, force=not update)
    write_cached_data(mixdb.location, "mixture", mixdb.mixture(mixid).name, results)
    write_mixture_metadata(mixdb, mixture=mixdb.mixture(mixid))

    return set(results.keys())


def main() -> None:
    from docopt import docopt

    from sonusai import __version__ as sai_version
    from sonusai.mixture import MixtureDatabase
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sai_version, options_first=True)

    verbose = args["--verbose"]
    mixids = args["--mixid"]
    includes = {x.strip() for x in args["--include"].replace(" ", ",").lower().split(",") if x != ""}
    excludes = {x.strip() for x in args["--exclude"].replace(" ", ",").lower().split(",") if x != ""}
    update = args["--update"]
    show_supported = args["--supported"]
    dryrun = args["--dryrun"]
    num_proc = args["--nproc"]
    location = args["LOC"]

    import fnmatch
    import sys
    import time
    from functools import partial
    from os.path import join

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.utils import par_track
    from sonusai.utils import seconds_to_hms
    from sonusai.utils import track

    start_time = time.monotonic()

    # Setup logging file
    create_file_handler(join(location, "genmetrics.log"), verbose)
    update_console_handler(verbose)
    initial_log_messages("genmetrics")

    logger.info(f"Load mixture database from {location}")

    mixdb = MixtureDatabase(location)
    supported = mixdb.supported_metrics
    if show_supported:
        logger.info(f"\nSupported metrics:\n\n{supported.pretty}")
        sys.exit(0)

    # Handle default excludes
    if "none" in excludes:
        if "all" in includes:
            excludes = {"*asr*", "*wer*"}
        else:
            excludes = set()

    # Handle default includes
    if "all" in includes:
        includes = {"*"}

    included_metrics: set[str] = set()
    for include in includes:
        for m in fnmatch.filter(supported.names, include):
            included_metrics.add(m)

    excluded_metrics: set[str] = set()
    for exclude in excludes:
        for m in fnmatch.filter(supported.names, exclude):
            excluded_metrics.add(m)

    requested = included_metrics - excluded_metrics

    metrics = sorted(requested)

    if len(metrics) == 0:
        logger.warning("No metrics were requested")
        sys.exit(1)

    logger.info("Generating metrics:")
    logger.info(f"{', '.join(metrics)}")
    if dryrun:
        sys.exit(0)

    mixids = mixdb.mixids_to_list(mixids)
    logger.info("")
    logger.info(f"Found {len(mixids):,} mixtures to process")

    if num_proc is None or len(mixids) == 1:
        no_par = True
        num_proc = None
    else:
        no_par = False
        num_proc = int(num_proc)  # TBD add support for 'auto'

    progress = track(total=len(mixids), desc="genmetrics")
    results = par_track(
        partial(_process_mixture, location=location, metrics=metrics, update=update),
        mixids,
        progress=progress,
        num_cpus=num_proc,
        no_par=no_par,
    )
    progress.close()

    written_metrics = sorted(set().union(*results))
    logger.info(f"Wrote metrics for {len(mixids)} mixtures to {location}:")
    logger.info(f"{', '.join(written_metrics)}")
    logger.info("")

    end_time = time.monotonic()
    logger.info(f"Completed in {seconds_to_hms(seconds=end_time - start_time)}")
    logger.info("")


if __name__ == "__main__":
    from sonusai import exception_handler
    from sonusai.utils import register_keyboard_interrupt

    register_keyboard_interrupt()
    try:
        main()
    except Exception as e:
        exception_handler(e)



from collections.abc import Callable

from ..datatypes import ClassCount
from ..mixture.helpers import mixture_all_speech_metadata
from ..mixture.mixdb import MixtureDatabase


def print_mixture_details(
    mixdb: MixtureDatabase,
    mixid: int | None = None,
    print_fn: Callable = print,
) -> None:
    from ..utils.seconds_to_hms import seconds_to_hms

    if mixid is not None:
        if 0 < mixid >= mixdb.num_mixtures:
            raise ValueError(f"Given mixid is outside valid range of 0:{mixdb.num_mixtures - 1}.")

        print_fn(f"Mixture {mixid} details")
        mixture = mixdb.mixture(mixid)
        speech_metadata = mixture_all_speech_metadata(mixdb, mixture)
        for category, source in mixture.all_sources.items():
            source_file = mixdb.source_file(source.file_id)
            print_fn(f"  {category}")
            print_fn(f"    name: {source_file.name}")
            print_fn(f"    effects: {source.effects.to_dict()}")
            print_fn(f"    pre_tempo: {source.pre_tempo}")
            print_fn(f"    duration: {seconds_to_hms(source_file.duration)}")
            print_fn(f"    start: {source.start}")
            print_fn(f"    repeat: {source.repeat}")
            print_fn(f"    snr: {source.snr}")
            print_fn(f"    random_snr: {source.snr.is_random}")
            print_fn(f"    snr_gain: {source.snr_gain}")
            for key in source_file.truth_configs:
                print_fn(f"    truth '{key}' function: {source_file.truth_configs[key].function}")
                print_fn(f"    truth '{key}' config: {source_file.truth_configs[key].config}")
                print_fn(
                    f"    truth '{key}' stride_reduction: {source_file.truth_configs[key].stride_reduction}"
                )
            for key in speech_metadata[category]:
                print_fn(f"{category} speech {key}: {speech_metadata[category][key]}")
        print_fn(f"  samples: {mixture.samples}")
        print_fn(f"  feature frames: {mixdb.mixture_feature_frames(mixid)}")
        print_fn("")


def print_class_count(
    class_count: ClassCount,
    length: int,
    print_fn: Callable = print,
    all_class_counts: bool = False,
) -> None:
    from ..utils.max_text_width import max_text_width

    print_fn("Class count:")
    idx_len = max_text_width(len(class_count))
    for idx, count in enumerate(class_count):
        if all_class_counts or count > 0:
            desc = f"  class {idx + 1:{idx_len}}"
            print_fn(f"{desc:{length}} {count}")



# SonusAI truth functions
# ruff: noqa: F401

from .crm import crm
from .crm import crm_parameters
from .crm import crm_validate
from .crm import crmp
from .crm import crmp_parameters
from .crm import crmp_validate
from .energy import energy_f
from .energy import energy_f_parameters
from .energy import energy_f_validate
from .energy import energy_t
from .energy import energy_t_parameters
from .energy import energy_t_validate
from .energy import mapped_snr_f
from .energy import mapped_snr_f_parameters
from .energy import mapped_snr_f_validate
from .energy import snr_f
from .energy import snr_f_parameters
from .energy import snr_f_validate
from .file import file
from .file import file_parameters
from .file import file_validate
from .metadata import metadata
from .metadata import metadata_parameters
from .metadata import metadata_validate
from .metrics import metrics
from .metrics import metrics_parameters
from .metrics import metrics_validate
from .phoneme import phoneme
from .phoneme import phoneme_parameters
from .phoneme import phoneme_validate
from .sed import sed
from .sed import sed_parameters
from .sed import sed_validate
from .target import target_f
from .target import target_f_parameters
from .target import target_f_validate
from .target import target_mixture_f
from .target import target_mixture_f_parameters
from .target import target_mixture_f_validate
from .target import target_swin_f
from .target import target_swin_f_parameters
from .target import target_swin_f_validate



from collections import namedtuple

SourceFileRecord = namedtuple(
    "SourceFileRecord",
    [
        "id",
        "category",
        "class_indices",
        "level_type",
        "name",
        "samples",
        "speaker_id",
    ],
)

TopRecord = namedtuple(
    "TopRecord",
    [
        "id",
        "class_balancing",
        "feature",
        "mixid_width",
        "num_classes",
        "seed",
        "speaker_metadata_tiers",
        "textgrid_metadata_tiers",
        "version",
    ],
)

ClassLabelRecord = namedtuple(
    "ClassLabelRecord",
    [
        "id",
        "label",
    ],
)

ClassWeightsThresholdRecord = namedtuple(
    "ClassWeightsThresholdRecord",
    [
        "id",
        "threshold",
    ],
)

ImpulseResponseFileRecord = namedtuple(
    "ImpulseResponseFileRecord",
    [
        "id",
        "delay",
        "name",
    ],
)

SpectralMaskRecord = namedtuple(
    "SpectralMaskRecord",
    [
        "id",
        "f_max_width",
        "f_num",
        "t_max_percent",
        "t_max_width",
        "t_num",
    ],
)

SourceRecord = namedtuple(
    "SourceRecord",
    [
        "id",
        "effects",
        "file_id",
        "pre_tempo",
        "repeat",
        "snr",
        "snr_gain",
        "snr_random",
        "start",
    ],
)

MixtureRecord = namedtuple(
    "MixtureRecord",
    [
        "id",
        "name",
        "samples",
        "spectral_mask_id",
        "spectral_mask_seed",
    ],
)



from collections.abc import Callable
from typing import Any

from ..datatypes import GeneralizedIDs
from ..mixture.mixdb import MixtureDatabase


def _true_predicate(_: Any) -> bool:
    return True


def get_mixids_from_mixture_field_predicate(
    mixdb: MixtureDatabase,
    field: str,
    mixids: GeneralizedIDs = "*",
    predicate: Callable[[Any], bool] | None = None,
) -> dict[int, list[int]]:
    """
    Generate mixture IDs based on mixture field and predicate
    Return a dictionary where:
        - keys are the matching field values
        - values are lists of the mixids that match the criteria
    """
    mixid_out = mixdb.mixids_to_list(mixids)

    if predicate is None:
        predicate = _true_predicate

    criteria_set = set()
    for m_id in mixid_out:
        value = getattr(mixdb.mixture(m_id), field)
        if isinstance(value, dict):
            for v in value.values():
                if predicate(v):
                    criteria_set.add(v)
        elif predicate(value):
            criteria_set.add(value)
    criteria = sorted(criteria_set)

    result: dict[int, list[int]] = {}
    for criterion in criteria:
        result[criterion] = []
        for m_id in mixid_out:
            value = getattr(mixdb.mixture(m_id), field)
            if isinstance(value, dict):
                for v in value.values():
                    if v == criterion:
                        result[criterion].append(m_id)
            elif value == criterion:
                result[criterion].append(m_id)

    return result


def get_mixids_from_truth_configs_field_predicate(
    mixdb: MixtureDatabase,
    field: str,
    mixids: GeneralizedIDs = "*",
    predicate: Callable[[Any], bool] | None = None,
) -> dict[int, list[int]]:
    """
    Generate mixture IDs based on target truth_configs field and predicate
    Return a dictionary where:
        - keys are the matching field values
        - values are lists of the mixids that match the criteria
    """
    from ..mixture.constants import REQUIRED_TRUTH_CONFIGS

    mixid_out = mixdb.mixids_to_list(mixids)

    # Get all field values
    values = get_all_truth_configs_values_from_field(mixdb, field)

    if predicate is None:
        predicate = _true_predicate

    # Get only values of interest
    values = [value for value in values if predicate(value)]

    result = {}
    for value in values:
        # Get a list of sources for each field value
        indices = []
        for s_ids in mixdb.source_file_ids.values():
            for s_id in s_ids:
                source = mixdb.source_file(s_id)
                for truth_config in source.truth_configs.values():
                    if field in REQUIRED_TRUTH_CONFIGS:
                        if value in getattr(truth_config, field):
                            indices.append(s_id)
                    else:
                        if value in getattr(truth_config.config, field):
                            indices.append(s_id)
        indices = sorted(set(indices))

        mixids = []
        for index in indices:
            for m_id in mixid_out:
                if index in [source.file_id for source in mixdb.mixture(m_id).all_sources.values()]:
                    mixids.append(m_id)

        mixids = sorted(set(mixids))
        if mixids:
            result[value] = mixids

    return result


def get_all_truth_configs_values_from_field(mixdb: MixtureDatabase, field: str) -> list:
    """
    Generate a list of all values corresponding to the given field in truth_configs
    """
    from ..mixture.constants import REQUIRED_TRUTH_CONFIGS

    result = []
    for sources in mixdb.source_files.values():
        for source in sources:
            for truth_config in source.truth_configs.values():
                if field in REQUIRED_TRUTH_CONFIGS:
                    value = getattr(truth_config, field)
                else:
                    value = getattr(truth_config.config, field, None)
                if not isinstance(value, list):
                    value = [value]
                result.extend(value)

    return sorted(set(result))


def get_mixids_from_noise(
    mixdb: MixtureDatabase,
    mixids: GeneralizedIDs = "*",
    predicate: Callable[[Any], bool] | None = None,
) -> dict[int, list[int]]:
    """
    Generate mixids based on noise index predicate
    Return a dictionary where:
        - keys are the noise indices
        - values are lists of the mixids that match the noise index
    """
    return get_mixids_from_mixture_field_predicate(mixdb=mixdb, mixids=mixids, field="noise_id", predicate=predicate)


def get_mixids_from_source(
    mixdb: MixtureDatabase,
    mixids: GeneralizedIDs = "*",
    predicate: Callable[[Any], bool] | None = None,
) -> dict[int, list[int]]:
    """
    Generate mixids based on a source index predicate
    Return a dictionary where:
        - keys are the source indices
        - values are lists of the mixids that match the source index
    """
    return get_mixids_from_mixture_field_predicate(mixdb=mixdb, mixids=mixids, field="source_ids", predicate=predicate)


def get_mixids_from_snr(
    mixdb: MixtureDatabase,
    mixids: GeneralizedIDs = "*",
    predicate: Callable[[Any], bool] | None = None,
) -> dict[float, list[int]]:
    """
    Generate mixids based on an SNR predicate
    Return a dictionary where:
        - keys are the SNRs
        - values are lists of the mixids that match the SNR
    """
    mixid_out = mixdb.mixids_to_list(mixids)

    # Get all the SNRs
    snrs = [float(snr) for snr in mixdb.all_snrs if not snr.is_random]

    if predicate is None:
        predicate = _true_predicate

    # Get only the SNRs of interest (filter on predicate)
    snrs = [snr for snr in snrs if predicate(snr)]

    result: dict[float, list[int]] = {}
    for snr in snrs:
        # Get a list of mixids for each SNR
        result[snr] = sorted(
            [i for i, mixture in enumerate(mixdb.mixtures) if mixture.noise.snr == snr and i in mixid_out]
        )

    return result


def get_mixids_from_class_indices(
    mixdb: MixtureDatabase,
    mixids: GeneralizedIDs = "*",
    predicate: Callable[[Any], bool] | None = None,
) -> dict[int, list[int]]:
    """
    Generate mixids based on a class index predicate
    Return a dictionary where:
        - keys are the class indices
        - values are lists of the mixids that match the class index
    """
    mixid_out = mixdb.mixids_to_list(mixids)

    if predicate is None:
        predicate = _true_predicate

    criteria_set = set()
    for m_id in mixid_out:
        class_indices = mixdb.mixture_class_indices(m_id)
        for class_index in class_indices:
            if predicate(class_index):
                criteria_set.add(class_index)
    criteria = sorted(criteria_set)

    result: dict[int, list[int]] = {}
    for criterion in criteria:
        result[criterion] = []
        for m_id in mixid_out:
            class_indices = mixdb.mixture_class_indices(m_id)
            for class_index in class_indices:
                if class_index == criterion:
                    result[criterion].append(m_id)

    return result


def get_mixids_from_truth_function(
    mixdb: MixtureDatabase,
    mixids: GeneralizedIDs = "*",
    predicate: Callable[[Any], bool] | None = None,
) -> dict[int, list[int]]:
    """
    Generate mixids based on a truth function predicate
    Return a dictionary where:
        - keys are the truth functions
        - values are lists of the mixids that match the truth function
    """
    return get_mixids_from_truth_configs_field_predicate(
        mixdb=mixdb, mixids=mixids, field="function", predicate=predicate
    )



def evaluate_random_rule(rule: str) -> str | float:
    """Evaluate 'rand' directive

    :param rule: Rule
    :return: Resolved value
    """
    import re
    from random import uniform

    rand_pattern = re.compile(r"rand\(([-+]?(\d+(\.\d*)?|\.\d+)),\s*([-+]?(\d+(\.\d*)?|\.\d+))\)")

    def rand_repl(m):
        return f"{uniform(float(m.group(1)), float(m.group(4))):.2f}"  # noqa: S311

    return eval(re.sub(rand_pattern, rand_repl, rule))  # noqa: S307



import numpy as np


def calc_sa_sdr(
    hypothesis: np.ndarray,
    reference: np.ndarray,
    with_scale: bool = False,
    with_negate: bool = False,
) -> tuple[np.ndarray, np.ndarray]:
    """Calculate source-aggregated SDR (signal distortion ratio) using all source inputs which are [samples, nsrc].

    These should include a noise to be a complete mixture estimate, i.e.,
        noise_est = sum-over-all-srcs(s_est(0:nsamples, :) - sum-over-non-noisesrc(s_est(0:nsamples, n))
    should be one of the sources in reference (s_true) and hypothesis (s_est).

    Calculates -10*log10(sumn(||sn||^2) / sumn(||sn - shn||^2)
    Note: for SA method, sums are done independently on ref and error before division, vs. SDR and SI-SDR
    where sum over n is taken after divide (before log).  This is more stable in noise-only cases and also
    when some sources are poorly estimated.
    TBD: add soft-max option with eps and tau params

    Reference:
        SA-SDR: A Novel Loss Function for Separation of Meeting Style Data
        Thilo von Neumann, Keisuke Kinoshita, Christoph Boeddeker, Marc Delcroix, Reinhold Haeb-Umbach
        https://doi.org/10.48550/arXiv.2110.15581

    :param hypothesis: [samples, nsrc]
    :param reference: [samples, nsrc]
    :param with_scale: enable scaling (scaling is same as in SI-SDR)
    :param with_negate: enable negation (for use as a loss function)
    :return: (sa_sdr, opt_scale)
    """
    if with_scale:
        # calc 1 x nsrc scaling factors
        ref_energy = np.sum(reference**2, axis=0, keepdims=True)
        # if ref_energy is zero, just set scaling to 1.0
        with np.errstate(divide="ignore", invalid="ignore"):
            opt_scale = np.sum(reference * hypothesis, axis=0, keepdims=True) / ref_energy
            opt_scale[opt_scale == np.inf] = 1.0
            opt_scale = np.nan_to_num(opt_scale, nan=1.0)
        scaled_ref = opt_scale * reference
    else:
        scaled_ref = reference
        opt_scale = np.ones((1, reference.shape[1]), dtype=float)

    # multisrc sa-sdr, inputs must be [samples, nsrc]
    err = scaled_ref - hypothesis

    # -10*log10(sumk(||sk||^2) / sumk(||sk - shk||^2)
    # sum over samples and sources
    num = np.sum(reference**2)
    den = np.sum(err**2)
    if num == 0 and den == 0:
        ratio = np.inf
    else:
        ratio = num / (den + np.finfo(np.float32).eps)

    sa_sdr = 10 * np.log10(ratio)

    if with_negate:
        # for use as a loss function
        sa_sdr = -sa_sdr

    return sa_sdr, opt_scale



from numpy.ma.core import choose
from urllib3.filepost import choose_boundary

from ..datatypes import AudioT
from ..datatypes import Effects
from .mixdb import MixtureDatabase


def get_effect_rules(location: str, config: dict, test: bool = False) -> dict[str, list[Effects]]:
    from ..datatypes import Effects
    from ..utils.dataclass_from_dict import list_dataclass_from_dict
    from .mixdb import MixtureDatabase

    mixdb = MixtureDatabase(location, test)

    rules: dict[str, list[Effects]] = {}
    for category, source in config["sources"].items():
        processed_rules: list[dict] = []
        for rule in source["effects"]:
            rule = _parse_ir_rule(rule, mixdb.num_ir_files)
            processed_rules = _expand_effect_rules(processed_rules, rule)
        rules[category] = list_dataclass_from_dict(list[Effects], processed_rules)

    validate_rules(mixdb, rules)
    return rules


def _expand_effect_rules(expanded_rules: list[dict], rule: dict) -> list[dict]:
    from copy import deepcopy

    from .parse import sai_expand

    for key in ("pre", "post"):
        if key in rule:
            value = rule[key]
            for idx in range(len(value)):
                new_rules = sai_expand(value[idx])
                if len(new_rules) > 1:
                    for new_rule in new_rules:
                        expanded_effect = deepcopy(rule)
                        new_value = deepcopy(value)
                        new_value[idx] = new_rule
                        expanded_effect[key] = new_value
                        _expand_effect_rules(expanded_rules, expanded_effect)
                    return expanded_rules

    expanded_rules.append(rule)
    return expanded_rules


def _parse_ir_rule(rule: dict, num_ir: int) -> dict:
    from ..datatypes import EffectList
    from .helpers import generic_ids_to_list

    def _resolve_str(parameters: str) -> str:
        if parameters.startswith("sai_"):
            return f"ir {parameters}"

        irs = generic_ids_to_list(num_ir, parameters)

        if not all(ro in range(num_ir) for ro in irs):
            raise ValueError(f"Invalid ir of {parameters}")

        if len(irs) == 1:
            return f"ir {irs[0]}"
        return f"ir sai_expand({', '.join(map(str, irs))})"

    def _process(rules_in: EffectList) -> EffectList:
        rules_out: EffectList = []

        for rule_in in rules_in:
            parts = rule_in.split(maxsplit=1)

            name = parts[0]
            if name != "ir":
                rules_out.append(rule_in)
                continue

            if len(parts) == 1:
                continue

            parameters = parts[1]
            if parameters.isnumeric():
                ir = int(parameters)
                if ir not in range(num_ir):
                    raise ValueError(f"Invalid ir of {parameters}")
                rules_out.append(rule_in)
                continue

            if isinstance(parameters, str):
                rules_out.append(_resolve_str(parameters))
                continue

            raise ValueError(f"Invalid ir of {parameters}")

        return rules_out

    for key in ("pre", "post"):
        if key in rule:
            rule[key] = _process(rule[key])

    return rule


def apply_effects(
    mixdb: MixtureDatabase,
    audio: AudioT,
    effects: Effects,
    pre: bool = True,
    post: bool = True,
) -> AudioT:
    """Apply effects to audio data

    :param mixdb: Mixture database
    :param audio: Input audio
    :param effects: Effects
    :param pre: Apply pre-truth effects
    :param post: Apply post-truth effects
    :return: Output audio
    """
    from ..datatypes import EffectList
    from .ir_effects import apply_ir
    from .ir_effects import read_ir
    from .sox_effects import apply_sox_effects

    def _process(audio_in: AudioT, effects_in) -> AudioT:
        _effects: EffectList = []
        for effect in effects_in:
            if effect.startswith("ir "):
                # Apply effects gathered so far
                audio_in = apply_sox_effects(audio_in, _effects)

                # Then empty the effects list
                _effects = []

                # Apply IR
                index = int(effect.split()[1])
                audio_in = apply_ir(
                    audio=audio_in,
                    ir=read_ir(
                        name=mixdb.ir_file(index),
                        delay=mixdb.ir_delay(index),
                        use_cache=mixdb.use_cache,
                    ),
                )
            else:
                _effects.append(effect)

        return apply_sox_effects(audio_in, _effects)

    audio_out = audio.copy()

    if pre:
        audio_out = _process(audio_out, effects.pre)

    if post:
        audio_out = _process(audio_out, effects.post)

    return audio_out


def estimate_effected_length(
    samples: int,
    effects: Effects,
    frame_length: int = 1,
    pre: bool = True,
    post: bool = True,
) -> int:
    """Estimate effected audio length

    :param samples: Original length in samples
    :param effects: Effects
    :param frame_length: Length will be a multiple of this
    :param pre: Apply pre-truth effects
    :param post: Apply post-truth effects
    :return: Estimated length in samples
    """
    from .pad_audio import get_padded_length

    def _update_samples(s: int, e: str) -> int:
        import re

        # speed factor[c]
        speed_pattern = re.compile(r"^speed\s+(-?\d+(\.\d+)*)(c?)$")
        result = re.search(speed_pattern, e)
        if result:
            value = float(result.group(1))
            if result.group(3):
                value = float(2 ** (value / 1200))
            return int(s / value + 0.5)

        # tempo [-q] [-m|-s|-l] factor [segment [search [overlap]]]
        tempo_pattern = re.compile(r"^tempo\s+(-q\s+)?(((-m)|(-s)|(-l))\s+)?(\d+(\.\d+)*)")
        result = re.search(tempo_pattern, e)
        if result:
            value = float(result.group(7))
            return int(s / value + 0.5)

        # other effects which do not affect length
        return s

    length = samples

    if pre:
        for effect in effects.pre:
            length = _update_samples(length, effect)

    if post:
        for effect in effects.post:
            length = _update_samples(length, effect)

    return get_padded_length(length, frame_length)


def evaluate_sai_random_float(text: str) -> str:
    """Evaluate 'sai_rand(min, max)' directive

    :param text: Text to evaluate
    :return: Resolved rule
    """
    import re
    from random import uniform

    def rand_repl(m):
        value = uniform(float(m.group(1)), float(m.group(4)))  # noqa: S311
        return f"{value:.2f}"

    rand_pattern = re.compile(r"sai_rand\(([-+]?(\d+(\.\d*)?|\.\d+)),\s*([-+]?(\d+(\.\d*)?|\.\d+))\)")

    resolved = text
    count = 0
    while re.findall(rand_pattern, resolved) and count < 100:
        try:
            resolved = re.sub(rand_pattern, rand_repl, resolved)
            count += 1
        except Exception as e:
            raise ValueError(f"Invalid rule: '{text}'.") from e

    if count == 100:
        raise ValueError(f"Invalid rule: '{text}'.")

    return resolved


def evaluate_sai_choose_ir(mixdb: MixtureDatabase, text: str) -> str:
    """Evaluate 'sai_choose' directive for ir

    :param mixdb: Mixture database
    :param text: Text to evaluate
    :return: Resolved value
    """
    import re
    from random import choice
    from random import randint

    choose_pattern = re.compile(r"^ir sai_choose\(\)$")
    choose_range_pattern = re.compile(r"^ir sai_choose\(([-+]?\d+),\s*([-+]?\d+)\)$")
    choose_tag_pattern = re.compile(r"^ir sai_choose\((\w+)\)$")

    def choose_range_repl(m) -> str:
        lower = int(m.group(1))
        upper = int(m.group(2))
        if (
            lower < 0
            or lower >= mixdb.num_ir_files
            or upper < 0
            or upper >= mixdb.num_ir_files
            or lower >= upper
            or str(lower) != m.group(1)
            or str(upper) != m.group(2)
        ):
            raise ValueError(f"Invalid rule: '{text}'. Values must be integers between 0 and {mixdb.num_ir_files - 1}.")
        return f"ir {randint(lower, upper)}"  # noqa: S311

    def choose_tag_repl(m) -> str:
        return m.group(1)

    if re.match(choose_pattern, text):
        return f"ir {randint(0, mixdb.num_ir_files - 1)}"  # noqa: S311

    if re.match(choose_range_pattern, text):
        try:
            return f"ir {eval(re.sub(choose_range_pattern, choose_range_repl, text))}"  # noqa: S307
        except Exception as e:
            raise ValueError(
                f"Invalid rule: '{text}'. Values must be integers between 0 and {mixdb.num_ir_files - 1}."
            ) from e

    if re.match(choose_tag_pattern, text):
        tag = re.sub(choose_tag_pattern, choose_tag_repl, text)
        if tag in mixdb.ir_tags:
            return f"ir {choice(mixdb.ir_file_ids_for_tag(tag))}"  # noqa: S311

        raise ValueError(f"Invalid rule: '{text}'. Tag, '{tag}', not found in database.")

    raise ValueError(f"Invalid rule: '{text}'.")


def effects_from_rules(mixdb: MixtureDatabase, rules: Effects) -> Effects:
    from copy import deepcopy

    effects = deepcopy(rules)
    for key in ("pre", "post"):
        entries = getattr(effects, key)
        for idx, entry in enumerate(entries):
            if entry.find("sai_rand") != -1:
                entries[idx] = evaluate_sai_random_float(entry)
            if entry.startswith("ir sai_choose"):
                entries[idx] = evaluate_sai_choose_ir(mixdb, entry)
        setattr(effects, key, entries)

    return effects


def conform_audio_to_length(audio: AudioT, length: int, repeat: bool, start: int) -> AudioT:
    """Conform audio to given length

    :param audio: Audio to conform
    :param length: Length of output
    :param repeat: Repeat samples or pad
    :param start: Starting sample offset
    :return: Conformed audio
    """
    import numpy as np

    if repeat:
        return np.take(audio, range(start, start + length), mode="wrap")

    end = length + start
    return np.pad(audio[start:], (0, end - len(audio)))


def validate_rules(mixdb: MixtureDatabase, rules: dict[str, list[Effects]]) -> None:
    from .sox_effects import validate_sox_effects

    for rule_list in rules.values():
        for rule in rule_list:
            sox_effects: list[str] = []
            effects = effects_from_rules(mixdb, rule)

            for effect in effects.pre:
                if not effect.startswith("ir"):
                    sox_effects.append(effect)

            for effect in effects.post:
                for check in ("speed", "tempo"):
                    if check in effect:
                        raise ValueError(f"'{check}' effect is not allowed in post-truth effect chain.")

                if not effect.startswith("ir"):
                    sox_effects.append(effect)

            validate_sox_effects(sox_effects)



from ..datatypes import AudioT
from ..datatypes import Feature


def get_feature_from_audio(audio: AudioT, feature_mode: str) -> Feature:
    """Apply forward transform and generate feature data from audio data

    :param audio: Time domain audio data [samples]
    :param feature_mode: Feature mode
    :return: Feature data [frames, strides, feature_parameters]
    """
    from pyaaware import FeatureGenerator

    from ..datatypes import TransformConfig
    from .helpers import forward_transform

    fg = FeatureGenerator(feature_mode=feature_mode)

    audio_f = forward_transform(
        audio=audio,
        config=TransformConfig(
            length=fg.ftransform_length,
            overlap=fg.ftransform_overlap,
            bin_start=fg.bin_start,
            bin_end=fg.bin_end,
            ttype=fg.ftransform_ttype,
        ),
    )

    return fg.execute_all(audio_f)[0]


def get_audio_from_feature(feature: Feature, feature_mode: str) -> AudioT:
    """Apply inverse transform to feature data to generate audio data

    :param feature: Feature data [frames, stride=1, feature_parameters]
    :param feature_mode: Feature mode
    :return: Audio data [samples]
    """
    import numpy as np
    from pyaaware import FeatureGenerator

    from ..datatypes import TransformConfig
    from ..utils.compress import power_uncompress
    from ..utils.stacked_complex import unstack_complex
    from .helpers import inverse_transform

    if feature.ndim != 3:
        raise ValueError("feature must have 3 dimensions: [frames, stride=1, feature_parameters]")

    if feature.shape[1] != 1:
        raise ValueError("Strided feature data is not supported for audio extraction; stride must be 1.")

    fg = FeatureGenerator(feature_mode=feature_mode)

    feature_complex = unstack_complex(feature.squeeze())
    if feature_mode[0:1] == "h":
        feature_complex = power_uncompress(feature_complex)
    return np.squeeze(
        inverse_transform(
            transform=feature_complex,
            config=TransformConfig(
                length=fg.itransform_length,
                overlap=fg.itransform_overlap,
                bin_start=fg.bin_start,
                bin_end=fg.bin_end,
                ttype=fg.itransform_ttype,
            ),
        )
    )



from functools import lru_cache
from pathlib import Path

from ..datatypes import AudioT
from ..datatypes import ImpulseResponseData
from .audio import raw_read_audio


def apply_ir(audio: AudioT, ir: ImpulseResponseData) -> AudioT:
    """Apply impulse response to audio data using scipy

    :param audio: Audio
    :param ir: Impulse response data
    :return: Effected audio
    """
    import numpy as np
    from librosa import resample
    from scipy.signal import fftconvolve

    from ..constants import SAMPLE_RATE

    if not isinstance(audio, np.ndarray):
        raise TypeError("audio must be a numpy array")

    # Early exit if no ir or if all audio is zero
    if ir is None or not audio.any():
        return audio

    pk_in = np.max(np.abs(audio))

    # Convert audio to IR sample rate
    audio_in = resample(audio, orig_sr=SAMPLE_RATE, target_sr=ir.sample_rate, res_type="soxr_hq")

    # Apply IR
    audio_out = fftconvolve(audio_in, ir.data, mode="full")

    # Delay compensation
    audio_out = audio_out[ir.delay :]

    # Convert back to global sample rate
    audio_out = resample(audio_out, orig_sr=ir.sample_rate, target_sr=SAMPLE_RATE, res_type="soxr_hq")

    # Trim to length
    audio_out = audio_out[: len(audio)]

    # Gain compensation
    pk_out = np.max(np.abs(audio_out))
    pk_gain = pk_in / pk_out
    audio_out = audio_out * pk_gain

    return audio_out


def read_ir(name: str | Path, delay: int, use_cache: bool = True) -> ImpulseResponseData:
    """Read impulse response data

    :param name: File name
    :param delay: Delay in samples
    :param use_cache: If true, use LRU caching
    :return: ImpulseResponseData object
    """
    if use_cache:
        return _read_ir(name, delay)
    return _read_ir.__wrapped__(name, delay)


@lru_cache
def _read_ir(name: str | Path, delay: int) -> ImpulseResponseData:
    """Read impulse response data using soundfile

    :param name: File name
    :param delay: Delay in samples
    :return: ImpulseResponseData object
    """
    out, sample_rate = raw_read_audio(name)

    return ImpulseResponseData(data=out, sample_rate=sample_rate, delay=delay)



import os
from dataclasses import dataclass
from pathlib import Path


@dataclass(frozen=True)
class Segment:
    person: str
    video: str
    id: str
    start: float
    stop: float


def load_speakers(input_dir: Path) -> dict:
    import csv

    speakers = {}

    # VoxCeleb1
    first = True
    with open(input_dir / "vox1_meta.csv", newline="") as file:
        data = csv.reader(file, delimiter="\t")
        for row in data:
            if first:
                first = False
            else:
                speakers[row[0].strip()] = {
                    "gender": row[2].strip(),
                    "dialect": row[3].strip(),
                    "category": row[4].strip(),
                }

    # VoxCeleb2
    first = True
    with open(input_dir / "vox2_meta.csv", newline="") as file:
        data = csv.reader(file, delimiter="\t")
        for row in data:
            if first:
                first = False
            else:
                speakers[row[1].strip()] = {
                    "gender": row[3].strip(),
                    "category": row[4].strip(),
                }

    return speakers


def load_segment(path: str | os.PathLike[str]) -> Segment:
    path = Path(path)

    with path.open("r") as file:
        segment = file.read().strip()

    header, frames = segment.split("\n\n")
    header_dict = _parse_header(header)
    start, stop = _get_segment_boundaries(frames)

    return Segment(
        person=header_dict["Identity"],
        video=header_dict["Reference"],
        id=path.stem,
        start=start,
        stop=stop,
    )


def _parse_header(header: str) -> dict:
    def _parse_line(line: str) -> tuple[str, str]:
        """Parse a line of header text into a dictionary.

        Header text has the following format:

        Identity  : \tid00017
        Reference : \t7t6lfzvVaTM
        Offset    : \t1
        FV Conf   : \t16.647\t(1)
        ASD Conf  : \t4.465

        """
        k, v = line.split("\t", maxsplit=1)
        k = k[:-2].strip()
        v = v.strip()
        return k, v

    return dict(_parse_line(line) for line in header.split("\n"))


def _get_segment_boundaries(frames: str) -> tuple[float, float]:
    """Get the start and stop points of the segment.

    Frames text has the following format:

    FRAME 	X 	Y 	W 	H
    000245 	0.392 	0.223 	0.253 	0.451
    ...
    000470 	0.359 	0.207 	0.260 	0.463

    """

    def _get_frame_seconds(line: str) -> float:
        frame = int(line.split("\t")[0])
        # YouTube is 25 FPS
        return frame / 25

    lines = frames.split("\n")
    return _get_frame_seconds(lines[1]), _get_frame_seconds(lines[-1])



def yes_or_no(question: str) -> bool:
    """Wait for yes or no input"""
    while True:
        reply = str(input(question + " (y/n)?: ")).lower().strip()
        if reply[:1] == "y":
            return True
        if reply[:1] == "n":
            return False



from ..constants import SAMPLE_RATE
from ..datatypes import AudioT


def write_audio(name: str, audio: AudioT, sample_rate: int = SAMPLE_RATE) -> None:
    """Write an audio file.

    To write multiple channels, use a 2D array of shape [channels, samples].
    The bits per sample and PCM/float are determined by the data type.

    """
    import torch
    import torchaudio

    data = torch.tensor(audio)

    if data.dim() == 1:
        data = torch.reshape(data, (1, data.shape[0]))
    if data.dim() != 2:
        raise ValueError("audio must be a 1D or 2D array")

    # Assuming data has more samples than channels, check if array needs to be transposed
    if data.shape[1] < data.shape[0]:
        data = torch.transpose(data, 0, 1)

    torchaudio.save(uri=name, src=data, sample_rate=sample_rate)



from collections.abc import Iterable
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import NamedTuple
from typing import SupportsIndex
from typing import TypeAlias

import numpy as np
import numpy.typing as npt
from dataclasses_json import DataClassJsonMixin
from praatio.utilities.constants import Interval

AudioT: TypeAlias = npt.NDArray[np.float32]

Truth: TypeAlias = Any
TruthDict: TypeAlias = dict[str, Truth]
TruthsDict: TypeAlias = dict[str, TruthDict]
Segsnr: TypeAlias = npt.NDArray[np.float32]

AudioF: TypeAlias = npt.NDArray[np.complex64]

EnergyT: TypeAlias = npt.NDArray[np.float32]
EnergyF: TypeAlias = npt.NDArray[np.float32]

Feature: TypeAlias = npt.NDArray[np.float32]

Predict: TypeAlias = npt.NDArray[np.float32]

# Json type defined to maintain compatibility with DataClassJsonMixin
Json: TypeAlias = dict | list | str | int | float | bool | None


class DataClassSonusAIMixin(DataClassJsonMixin):
    def __str__(self):
        return f"{self.to_dict()}"

    # Override DataClassJsonMixin to remove dictionary keys with values of None
    def to_dict(self, encode_json=False) -> dict[str, Json]:
        def del_none(d):
            if isinstance(d, dict):
                for key, value in list(d.items()):
                    if value is None:
                        del d[key]
                    elif isinstance(value, dict):
                        del_none(value)
                    elif isinstance(value, list):
                        for item in value:
                            del_none(item)
            elif isinstance(d, list):
                for item in d:
                    del_none(item)
            return d

        return del_none(super().to_dict(encode_json))


@dataclass(frozen=True)
class TruthConfig(DataClassSonusAIMixin):
    function: str
    stride_reduction: str
    config: dict = field(default_factory=dict)

    def __hash__(self):
        return hash(self.to_json())

    def __eq__(self, other):
        return isinstance(other, TruthConfig) and hash(self) == hash(other)


TruthConfigs: TypeAlias = dict[str, TruthConfig]
TruthsConfigs: TypeAlias = dict[str, TruthConfigs]


NumberStr: TypeAlias = float | int | str
OptionalNumberStr: TypeAlias = NumberStr | None
OptionalListNumberStr: TypeAlias = list[NumberStr] | None


EffectList: TypeAlias = list[str]


@dataclass
class Effects(DataClassSonusAIMixin):
    pre: EffectList
    post: EffectList = field(default_factory=EffectList)


class UniversalSNRGenerator:
    def __init__(self, raw_value: float | str) -> None:
        self._raw_value = str(raw_value)
        self.is_random = isinstance(raw_value, str) and raw_value.startswith("sai_rand")

    @property
    def value(self) -> float:
        from sonusai.mixture.effects import evaluate_sai_random_float

        if self.is_random:
            return float(evaluate_sai_random_float(self._raw_value))
        return float(self._raw_value)


class UniversalSNR(float):
    def __new__(cls, value: float, is_random: bool = False):
        return float.__new__(cls, value)

    def __init__(self, value: float, is_random: bool = False) -> None:
        float.__init__(value)
        self._is_random = bool(is_random)

    @property
    def is_random(self) -> bool:
        return self._is_random


Speaker: TypeAlias = dict[str, str]


@dataclass
class SourceFile(DataClassSonusAIMixin):
    category: str
    class_indices: list[int]
    name: str
    samples: int
    truth_configs: TruthConfigs
    class_balancing_effect: EffectList | None = None
    id: int = -1
    level_type: str | None = None
    speaker_id: int | None = None

    @property
    def duration(self) -> float:
        from .constants import SAMPLE_RATE

        return self.samples / SAMPLE_RATE


@dataclass
class EffectedFile(DataClassSonusAIMixin):
    file_id: int
    effect_id: int


ClassCount: TypeAlias = list[int]

GeneralizedIDs: TypeAlias = str | int | list[int] | range


@dataclass(frozen=True)
class SpectralMask(DataClassSonusAIMixin):
    f_max_width: int
    f_num: int
    t_max_width: int
    t_num: int
    t_max_percent: int


@dataclass(frozen=True)
class TruthParameter(DataClassSonusAIMixin):
    category: str
    name: str
    parameters: int | None


@dataclass
class Source(DataClassSonusAIMixin):
    effects: Effects
    file_id: int
    pre_tempo: float = 1
    repeat: bool = False
    snr: UniversalSNR = field(default_factory=lambda: UniversalSNR(0))
    snr_gain: float = 0
    start: int = 0


Sources: TypeAlias = dict[str, Source]
SourcesAudioT: TypeAlias = dict[str, AudioT]
SourcesAudioF: TypeAlias = dict[str, AudioF]


@dataclass
class Mixture(DataClassSonusAIMixin):
    name: str
    samples: int
    all_sources: Sources
    spectral_mask_id: int
    spectral_mask_seed: int

    @property
    def all_source_ids(self) -> dict[str, int]:
        return {category: source.file_id for category, source in self.all_sources.items()}

    @property
    def sources(self) -> Sources:
        return {category: source for category, source in self.all_sources.items() if category != "noise"}

    @property
    def source_ids(self) -> dict[str, int]:
        return {category: source.file_id for category, source in self.sources.items()}

    @property
    def noise(self) -> Source:
        return self.all_sources["noise"]

    @property
    def noise_id(self) -> int:
        return self.noise.file_id

    @property
    def source_effects(self) -> dict[str, Effects]:
        return {category: source.effects for category, source in self.sources.items()}

    @property
    def noise_effects(self) -> Effects:
        return self.noise.effects

    @property
    def is_noise_only(self) -> bool:
        return self.noise.snr < -96

    @property
    def is_source_only(self) -> bool:
        return self.noise.snr > 96


@dataclass(frozen=True)
class TransformConfig:
    length: int
    overlap: int
    bin_start: int
    bin_end: int
    ttype: str


@dataclass(frozen=True)
class FeatureGeneratorConfig:
    feature_mode: str
    truth_parameters: dict[str, dict[str, int | None]]


@dataclass(frozen=True)
class FeatureGeneratorInfo:
    decimation: int
    stride: int
    step: int
    feature_parameters: int
    ft_config: TransformConfig
    eft_config: TransformConfig
    it_config: TransformConfig


ASRConfigs: TypeAlias = dict[str, dict[str, Any]]


@dataclass
class GenMixData:
    mixture: AudioT | None = None
    truth_t: TruthsDict | None = None
    segsnr_t: Segsnr | None = None
    sources: SourcesAudioT | None = None
    source: AudioT | None = None
    noise: AudioT | None = None


@dataclass
class GenFTData:
    feature: Feature | None = None
    truth_f: TruthsDict | None = None
    segsnr: Segsnr | None = None


@dataclass
class ImpulseResponseData:
    data: AudioT
    sample_rate: int
    delay: int


@dataclass
class ImpulseResponseFile(DataClassSonusAIMixin):
    name: str
    tags: list[str]
    delay: str | int = "auto"


@dataclass
class MixtureDatabaseConfig(DataClassSonusAIMixin):
    asr_configs: ASRConfigs
    class_balancing: bool
    class_labels: list[str]
    class_weights_threshold: list[float]
    feature: str
    ir_files: list[ImpulseResponseFile]
    mixtures: list[Mixture]
    num_classes: int
    source_files: dict[str, list[SourceFile]]
    spectral_masks: list[SpectralMask]


SpeechMetadata: TypeAlias = str | list[Interval] | None


class SnrFMetrics(NamedTuple):
    avg: float | None = None
    std: float | None = None
    db_avg: float | None = None
    db_std: float | None = None


class SnrFBinMetrics(NamedTuple):
    avg: np.ndarray | None = None
    std: np.ndarray | None = None
    db_avg: np.ndarray | None = None
    db_std: np.ndarray | None = None


class SpeechMetrics(NamedTuple):
    csig: float | None = None
    cbak: float | None = None
    covl: float | None = None


class AudioStatsMetrics(NamedTuple):
    dco: float | None = None
    min: float | None = None
    max: float | None = None
    pkdb: float | None = None
    lrms: float | None = None
    pkr: float | None = None
    tr: float | None = None
    cr: float | None = None
    fl: float | None = None
    pkc: float | None = None


@dataclass
class MetricDoc:
    category: str
    name: str
    description: str


class MetricDocs(list[MetricDoc]):
    def __init__(self, __iterable: Iterable[MetricDoc]) -> None:
        super().__init__(item for item in __iterable)

    def __setitem__(self, __key: SupportsIndex, __value: MetricDoc) -> None:  # type: ignore[override]
        super().__setitem__(__key, __value)

    def insert(self, __index: SupportsIndex, __object: MetricDoc) -> None:
        super().insert(__index, __object)

    def append(self, __object: MetricDoc) -> None:
        super().append(__object)

    def extend(self, __iterable: Iterable[MetricDoc]) -> None:
        if isinstance(__iterable, type(self)):
            super().extend(__iterable)
        else:
            super().extend(item for item in __iterable)

    @property
    def pretty(self) -> str:
        max_category_len = ((max([len(item.category) for item in self]) + 9) // 10) * 10
        max_name_len = 2 + ((max([len(item.name) for item in self]) + 1) // 2) * 2
        categories: list[str] = []
        for item in self:
            if item.category not in categories:
                categories.append(item.category)

        result = ""
        for category in categories:
            result += f"{category}\n"
            result += "-" * max_category_len + "\n"
            for item in [sub for sub in self if sub.category == category]:
                result += f"  {item.name:<{max_name_len}}{item.description}\n"
            result += "\n"

        return result

    @property
    def names(self) -> set[str]:
        return {item.name for item in self}



from ..datatypes import EffectList
from ..datatypes import EffectedFile
from ..datatypes import File


def balance_sources(
    effected_sources: list[EffectedFile],
    files: list[File],
    effects: list[EffectList],
    class_balancing_effect: EffectList,
    num_classes: int,
    num_ir: int,
    mixups: list[int] | None = None,
) -> tuple[list[EffectedFile], list[EffectList]]:
    import math

    from .augmentation import get_mixups
    from .sources import get_augmented_target_ids_by_class

    first_cba_id = len(effects)

    if mixups is None:
        mixups = get_mixups(effects)

    for mixup in mixups:
        if mixup == 1:
            continue

        effected_sources_indices_by_class = get_augmented_target_ids_by_class(
            augmented_targets=effected_sources,
            targets=files,
            target_augmentations=effects,
            mixup=mixup,
            num_classes=num_classes,
        )

        largest = max([len(item) for item in effected_sources_indices_by_class])
        largest = math.ceil(largest / mixup) * mixup
        for es_indices in effected_sources_indices_by_class:
            additional_effects_needed = largest - len(es_indices)
            file_ids = sorted({effected_sources[at_index].file_id for at_index in es_indices})

            tfi_idx = 0
            for _ in range(additional_effects_needed):
                file_id = file_ids[tfi_idx]
                tfi_idx = (tfi_idx + 1) % len(file_ids)
                effect_id, effects = _get_unused_balancing_effect(
                    effected_sources=effected_sources,
                    files=files,
                    effects=effects,
                    class_balancing_effect=class_balancing_effect,
                    file_id=file_id,
                    mixup=mixup,
                    num_ir=num_ir,
                    first_cbe_id=first_cba_id,
                )
                effected_sources.append(EffectedFile(file_id=file_id, effect_id=effect_id))

    return effected_sources, effects


def _get_unused_balancing_effect(
    effected_sources: list[EffectedFile],
    files: list[File],
    effects: list[EffectList],
    class_balancing_effect: EffectList,
    file_id: int,
    mixup: int,
    num_ir: int,
    first_cbe_id: int,
) -> tuple[int, list[EffectList]]:
    """Get an unused balancing augmentation for a given target file index"""
    from dataclasses import asdict

    from .augmentation import get_augmentation_rules

    balancing_augmentations = [item for item in range(len(effects)) if item >= first_cbe_id]
    used_balancing_augmentations = [
        effected_source.effect_id
        for effected_source in effected_sources
        if effected_source.file_id == file_id and effected_source.effect_id in balancing_augmentations
    ]

    augmentation_indices = [
        item
        for item in balancing_augmentations
        if item not in used_balancing_augmentations and effects[item].mixup == mixup
    ]
    if len(augmentation_indices) > 0:
        return augmentation_indices[0], effects

    class_balancing_effect = get_class_balancing_effect(file=files[file_id], default_cbe=class_balancing_effect)
    new_effect = get_augmentation_rules(rules=asdict(class_balancing_effect), num_ir=num_ir)[0]
    new_effect.mixup = mixup
    effects.append(new_effect)
    return len(effects) - 1, effects


def get_class_balancing_effect(file: File, default_cbe: EffectList) -> EffectList:
    """Get the class balancing effect rule for the given target"""
    if file.class_balancing_effect is not None:
        return file.class_balancing_effect
    return default_cbe



def trim_docstring(docstring: str | None) -> str:
    """Trim whitespace from docstring"""
    from sys import maxsize

    if not docstring:
        return ""

    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()

    # Determine minimum indentation (first line doesn't count)
    indent = maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))

    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())

    # Strip off leading blank lines:
    while trimmed and not trimmed[0]:
        trimmed.pop(0)

    # Return a single string
    return "\n".join(trimmed)


def add_commands_to_docstring(docstring: str | None, plugin_docstrings: list[str]) -> str:
    """Add commands to docstring"""
    import sonusai

    if not docstring:
        lines = []
    else:
        lines = docstring.splitlines()

    start = lines.index("The sonusai commands are:")
    end = lines.index("", start)

    commands = sonusai.commands_doc.splitlines()
    for plugin_docstring in plugin_docstrings:
        commands.extend(plugin_docstring.splitlines())
    commands.sort()
    commands = list(filter(None, commands))

    lines = lines[: start + 1] + commands + lines[end:]

    return "\n".join(lines)



def register_keyboard_interrupt() -> None:
    import signal

    def signal_handler(_sig, _frame):
        import sys

        from sonusai import logger

        logger.info("Canceled due to keyboard interrupt")
        sys.exit(1)

    signal.signal(signal.SIGINT, signal_handler)



"""sonusai genmixdb

usage: genmixdb [-hvmdjn] LOC

options:
    -h, --help
    -v, --verbose   Be verbose.
    -m, --mix       Save mixture data. [default: False].
    -d, --dryrun    Perform a dry run showing the processed config. [default: False].
    -j, --json      Save JSON version of database. [default: False].
    -n, --nopar     Do not run in parallel. [default: False].

Create mixture database data for training and evaluation. Optionally, also create mixture audio and feature/truth data.

genmixdb creates a database of training and evaluation feature and truth data generation information. It allows the
choice of audio neural-network feature types that are supported by the Aaware real-time front-end and truth data that is
synchronized frame-by-frame with the feature data.

For details, see sonusai doc.

"""

from sonusai.datatypes import Mixture
from sonusai.mixture import MixtureDatabase


def genmixdb(
    location: str,
    save_mix: bool = False,
    logging: bool = True,
    show_progress: bool = False,
    test: bool = False,
    verbose: bool = False,
    save_json: bool = False,
    no_par: bool = False,
) -> None:
    from functools import partial
    from random import seed

    import pandas as pd
    import yaml

    from sonusai import logger
    from sonusai.constants import SAMPLE_BYTES
    from sonusai.constants import SAMPLE_RATE
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import generate_mixtures
    from sonusai.mixture import get_effect_rules
    from sonusai.mixture import get_ir_files
    from sonusai.mixture import get_source_files
    from sonusai.mixture import initialize_db
    from sonusai.mixture import load_config
    from sonusai.mixture import log_duration_and_sizes
    from sonusai.mixture import populate_class_label_table
    from sonusai.mixture import populate_class_weights_threshold_table
    from sonusai.mixture import populate_impulse_response_file_table
    from sonusai.mixture import populate_mixture_table
    from sonusai.mixture import populate_source_file_table
    from sonusai.mixture import populate_spectral_mask_table
    from sonusai.mixture import populate_top_table
    from sonusai.mixture import populate_truth_parameters_table
    from sonusai.mixture import update_mixid_width
    from sonusai.utils import human_readable_size
    from sonusai.utils import par_track
    from sonusai.utils import seconds_to_hms
    from sonusai.utils import track

    config = load_config(location)
    initialize_db(location, test)

    mixdb = MixtureDatabase(location, test)

    populate_top_table(location, config, test, verbose)
    populate_class_label_table(location, config, test, verbose)
    populate_class_weights_threshold_table(location, config, test, verbose)
    populate_spectral_mask_table(location, config, test, verbose)
    populate_truth_parameters_table(location, config, test, verbose)

    seed(config["seed"])

    if logging:
        logger.debug(f"Seed: {config['seed']}")
        logger.debug("Configuration:")
        logger.debug(yaml.dump(config))

    if logging:
        logger.info("Collecting sources")

    source_files = get_source_files(config, show_progress)
    logger.info("")

    if len([file for file in source_files if file.category == "primary"]) == 0:
        raise RuntimeError("Canceled due to no primary sources")

    if logging:
        logger.info("Populating source file table")

    populate_source_file_table(location, source_files, test, verbose)

    if logging:
        logger.info("Sources summary")
        data = {
            "category": [],
            "files": [],
            "size": [],
            "duration": [],
        }
        for category, source_files in mixdb.source_files.items():
            audio_samples = sum([source.samples for source in source_files])
            audio_duration = audio_samples / SAMPLE_RATE
            data["category"].append(category)
            data["files"].append(mixdb.num_source_files(category))
            data["size"].append(human_readable_size(audio_samples * SAMPLE_BYTES, 1))
            data["duration"].append(seconds_to_hms(seconds=audio_duration))

        df = pd.DataFrame(data)
        logger.info(df.to_string(index=False, header=False))
        logger.info("")

        for category, files in mixdb.source_files.items():
            logger.debug(f"List of {category} sources:")
            logger.debug(yaml.dump([file.name for file in files], default_flow_style=False))

    if logging:
        logger.info("Collecting impulse responses")

    ir_files = get_ir_files(config, show_progress=show_progress)
    logger.info("")

    if logging:
        logger.info("Populating impulse response file table")

    populate_impulse_response_file_table(location, ir_files, test, verbose)

    if logging:
        logger.debug("List of impulse responses:")
        for idx, file in enumerate(ir_files):
            logger.debug(f"id: {idx}, name:{file.name}, delay: {file.delay}, tags: [{', '.join(file.tags)}]")
        logger.debug("")

    if logging:
        logger.info("Collecting effects")

    rules = get_effect_rules(location, config, test)

    if logging:
        logger.info("")
        for category, effect in rules.items():
            logger.debug(f"List of {category} rules:")
            logger.debug(yaml.dump([entry.to_dict() for entry in effect], default_flow_style=False))

    if logging:
        logger.debug("SNRS:")
        for category, source in config["sources"].items():
            if category != "primary":
                logger.debug(f"  {category}")
                for snr in source["snrs"]:
                    logger.debug(f"  - {snr}")
        logger.debug("")
        logger.debug("Mix Rules:")
        for category, source in config["sources"].items():
            if category != "primary":
                logger.debug(f"  {category}")
                for mix_rule in source["mix_rules"]:
                    logger.debug(f"  - {mix_rule}")
        logger.debug("")
        logger.debug("Spectral masks:")
        for spectral_mask in mixdb.spectral_masks:
            logger.debug(f"- {spectral_mask}")
        logger.debug("")

    if logging:
        logger.info("Generating mixtures")

    mixtures = generate_mixtures(location, config, rules, test)

    num_mixtures = len(mixtures)
    update_mixid_width(location, num_mixtures, test)

    if logging:
        logger.info(f"Found {num_mixtures:,} mixtures to process")

    total_duration = float(sum([mixture.samples for mixture in mixtures])) / SAMPLE_RATE

    if logging:
        log_duration_and_sizes(
            total_duration=total_duration,
            feature_step_samples=mixdb.feature_step_samples,
            feature_parameters=mixdb.feature_parameters,
            stride=mixdb.fg_stride,
            desc="Estimated",
        )
        logger.info(
            f"Feature shape:        "
            f"{mixdb.fg_stride} x {mixdb.feature_parameters} "
            f"({mixdb.fg_stride * mixdb.feature_parameters} total parameters)"
        )
        logger.info(f"Feature samples:      {mixdb.feature_samples} samples ({mixdb.feature_ms} ms)")
        logger.info(f"Feature step samples: {mixdb.feature_step_samples} samples ({mixdb.feature_step_ms} ms)")
        logger.info("")

    # Fill in the details
    if logging:
        logger.info("Processing mixtures")
    progress = track(total=num_mixtures, disable=not show_progress)
    mixtures = par_track(
        partial(
            _process_mixture,
            location=location,
            save_mix=save_mix,
            test=test,
        ),
        mixtures,
        progress=progress,
        no_par=no_par,
    )
    progress.close()

    populate_mixture_table(
        location=location,
        mixtures=mixtures,
        test=test,
        verbose=verbose,
        logging=logging,
        show_progress=show_progress,
    )

    total_duration = float(mixdb.total_samples() / SAMPLE_RATE)

    if logging:
        log_duration_and_sizes(
            total_duration=total_duration,
            feature_step_samples=mixdb.feature_step_samples,
            feature_parameters=mixdb.feature_parameters,
            stride=mixdb.fg_stride,
            desc="Actual",
        )
        logger.info("")

    if not test and save_json:
        if logging:
            logger.info(f"Writing JSON version of database to {location}")
        mixdb = MixtureDatabase(location)
        mixdb.save()


def _process_mixture(
    mixture: Mixture,
    location: str,
    save_mix: bool,
    test: bool,
) -> Mixture:
    from functools import partial

    from sonusai.mixture import update_mixture
    from sonusai.mixture import write_cached_data
    from sonusai.mixture import write_mixture_metadata

    mixdb = MixtureDatabase(location, test=test)
    mixture, genmix_data = update_mixture(mixdb, mixture, save_mix)

    write = partial(write_cached_data, location=location, name="mixture", index=mixture.name)

    if save_mix:
        write(
            items={
                "sources": genmix_data.sources,
                "source": genmix_data.source,
                "noise": genmix_data.noise,
                "mixture": genmix_data.mixture,
            }
        )

        write_mixture_metadata(mixdb, mixture=mixture)

    return mixture


def main() -> None:
    from docopt import docopt

    from sonusai import __version__ as sai_version
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sai_version, options_first=True)

    import time
    from os import makedirs
    from os import remove
    from os.path import exists
    from os.path import isdir
    from os.path import join

    import yaml

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.mixture import load_config
    from sonusai.utils import seconds_to_hms

    verbose = args["--verbose"]
    save_mix = args["--mix"]
    dryrun = args["--dryrun"]
    save_json = args["--json"]
    no_par = args["--nopar"]
    location = args["LOC"]

    start_time = time.monotonic()

    if exists(location) and not isdir(location):
        remove(location)

    makedirs(location, exist_ok=True)

    create_file_handler(join(location, "genmixdb.log"), verbose)
    update_console_handler(verbose)
    initial_log_messages("genmixdb")

    if dryrun:
        config = load_config(location)
        logger.info("Dryrun configuration:")
        logger.info(yaml.dump(config))
        return

    logger.info(f"Creating mixture database for {location}")
    logger.info("")

    genmixdb(
        location=location,
        save_mix=save_mix,
        show_progress=True,
        save_json=save_json,
        verbose=verbose,
        no_par=no_par,
    )

    end_time = time.monotonic()
    logger.info(f"Completed in {seconds_to_hms(seconds=end_time - start_time)}")
    logger.info("")


if __name__ == "__main__":
    from sonusai import exception_handler
    from sonusai.utils import register_keyboard_interrupt

    register_keyboard_interrupt()
    try:
        main()
    except Exception as e:
        exception_handler(e)



# ruff: noqa: S608
from functools import cached_property
from functools import lru_cache
from functools import partial
from typing import Any

from ..datatypes import ASRConfigs
from ..datatypes import AudioF
from ..datatypes import AudioT
from ..datatypes import ClassCount
from ..datatypes import Feature
from ..datatypes import FeatureGeneratorConfig
from ..datatypes import FeatureGeneratorInfo
from ..datatypes import GeneralizedIDs
from ..datatypes import ImpulseResponseFile
from ..datatypes import MetricDoc
from ..datatypes import MetricDocs
from ..datatypes import Mixture
from ..datatypes import Segsnr
from ..datatypes import SourceFile
from ..datatypes import Sources
from ..datatypes import SourcesAudioF
from ..datatypes import SourcesAudioT
from ..datatypes import SpectralMask
from ..datatypes import SpeechMetadata
from ..datatypes import TransformConfig
from ..datatypes import TruthConfigs
from ..datatypes import TruthDict
from ..datatypes import TruthsConfigs
from ..datatypes import TruthsDict
from ..datatypes import UniversalSNR
from .db import SQLiteDatabase
from .db import db_file


class MixtureDatabase:
    def __init__(self, location: str, test: bool = False, verbose: bool = False, use_cache: bool = True) -> None:
        import json
        from os.path import exists

        from .config import load_config

        self.location = location
        self.test = test
        self.verbose = verbose
        self.use_cache = use_cache

        if not exists(db_file(self.location, self.test)):
            raise OSError(f"Could not find mixture database in {self.location}")

        self.db = partial(SQLiteDatabase, location=self.location, test=self.test, verbose=self.verbose)

        # Check config.yml to see if asr_configs has changed and update the database if needed
        config = load_config(self.location)
        new_asr_configs = json.dumps(config["asr_configs"])
        with self.db() as c:
            old_asr_configs = c.execute("SELECT asr_configs FROM top").fetchone()

        if old_asr_configs is not None and new_asr_configs != old_asr_configs[0]:
            with SQLiteDatabase(
                location=self.location,
                readonly=False,
                test=self.test,
                verbose=self.verbose,
            ) as cur:
                cur.execute("UPDATE top SET asr_configs = ? WHERE ? = id", (new_asr_configs,))

    @cached_property
    def json(self) -> str:
        from ..datatypes import MixtureDatabaseConfig

        config = MixtureDatabaseConfig(
            asr_configs=self.asr_configs,
            class_balancing=self.class_balancing,
            class_labels=self.class_labels,
            class_weights_threshold=self.class_weights_thresholds,
            feature=self.feature,
            ir_files=self.ir_files,
            mixtures=self.mixtures,
            num_classes=self.num_classes,
            spectral_masks=self.spectral_masks,
            source_files=self.source_files,
        )
        return config.to_json(indent=2)

    def save(self) -> None:
        """Save the MixtureDatabase as a JSON file"""
        from os.path import join

        json_name = join(self.location, "mixdb.json")
        with open(file=json_name, mode="w") as file:
            file.write(self.json)

    @cached_property
    def fg_config(self) -> FeatureGeneratorConfig:
        return FeatureGeneratorConfig(
            feature_mode=self.feature,
            truth_parameters=self.truth_parameters,
        )

    @cached_property
    def fg_info(self) -> FeatureGeneratorInfo:
        from .helpers import get_feature_generator_info

        return get_feature_generator_info(self.fg_config)

    @cached_property
    def truth_parameters(self) -> dict[str, dict[str, int | None]]:
        with self.db() as c:
            rows = c.execute("SELECT category, name, parameters FROM truth_parameters").fetchall()
            truth_parameters: dict[str, dict[str, int | None]] = {}
            for row in rows:
                category, name, parameters = row
                if category not in truth_parameters:
                    truth_parameters[category] = {}
                truth_parameters[category][name] = parameters
            return truth_parameters

    @cached_property
    def num_classes(self) -> int:
        with self.db() as c:
            return int(c.execute("SELECT num_classes FROM top").fetchone()[0])

    @cached_property
    def asr_configs(self) -> ASRConfigs:
        import json

        with self.db() as c:
            return json.loads(c.execute("SELECT asr_configs FROM top").fetchone()[0])

    @cached_property
    def supported_metrics(self) -> MetricDocs:
        metrics = MetricDocs(
            [
                MetricDoc("Mixture Metrics", "mxsnr", "SNR specification in dB"),
                MetricDoc(
                    "Mixture Metrics",
                    "mxssnr_avg",
                    "Segmental SNR average over all frames",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxssnr_std",
                    "Segmental SNR standard deviation over all frames",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxssnrdb_avg",
                    "Segmental SNR average of the dB frame values over all frames",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxssnrdb_std",
                    "Segmental SNR standard deviation of the dB frame values over all frames",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxssnrf_avg",
                    "Per-bin segmental SNR average over all frames (using feature transform)",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxssnrf_std",
                    "Per-bin segmental SNR standard deviation over all frames (using feature transform)",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxssnrdbf_avg",
                    "Per-bin segmental average of the dB frame values over all frames (using feature transform)",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxssnrdbf_std",
                    "Per-bin segmental standard deviation of the dB frame values over all frames (using feature transform)",
                ),
                MetricDoc("Mixture Metrics", "mxpesq", "PESQ of mixture versus true sources"),
                MetricDoc(
                    "Mixture Metrics",
                    "mxwsdr",
                    "Weighted signal distortion ratio of mixture versus true sources",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxpd",
                    "Phase distance between mixture and true sources",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxstoi",
                    "Short term objective intelligibility of mixture versus true sources",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxcsig",
                    "Predicted rating of speech distortion of mixture versus true sources",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxcbak",
                    "Predicted rating of background distortion of mixture versus true sources",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxcovl",
                    "Predicted rating of overall quality of mixture versus true sources",
                ),
                MetricDoc("Mixture Metrics", "ssnr", "Segmental SNR"),
                MetricDoc("Mixture Metrics", "mxdco", "Mixture DC offset"),
                MetricDoc("Mixture Metrics", "mxmin", "Mixture min level"),
                MetricDoc("Mixture Metrics", "mxmax", "Mixture max levl"),
                MetricDoc("Mixture Metrics", "mxpkdb", "Mixture Pk lev dB"),
                MetricDoc("Mixture Metrics", "mxlrms", "Mixture RMS lev dB"),
                MetricDoc("Mixture Metrics", "mxpkr", "Mixture RMS Pk dB"),
                MetricDoc("Mixture Metrics", "mxtr", "Mixture RMS Tr dB"),
                MetricDoc("Mixture Metrics", "mxcr", "Mixture Crest factor"),
                MetricDoc("Mixture Metrics", "mxfl", "Mixture Flat factor"),
                MetricDoc("Mixture Metrics", "mxpkc", "Mixture Pk count"),
                MetricDoc("Sources Metrics", "sdco", "Sources DC offset"),
                MetricDoc("Sources Metrics", "smin", "Sources min level"),
                MetricDoc("Sources Metrics", "smax", "Sources max levl"),
                MetricDoc("Sources Metrics", "spkdb", "Sources Pk lev dB"),
                MetricDoc("Sources Metrics", "slrms", "Sources RMS lev dB"),
                MetricDoc("Sources Metrics", "spkr", "Sources RMS Pk dB"),
                MetricDoc("Sources Metrics", "str", "Sources RMS Tr dB"),
                MetricDoc("Sources Metrics", "scr", "Sources Crest factor"),
                MetricDoc("Sources Metrics", "sfl", "Sources Flat factor"),
                MetricDoc("Sources Metrics", "spkc", "Sources Pk count"),
                MetricDoc("Source Metrics", "mxsdco", "Source DC offset"),
                MetricDoc("Source Metrics", "mxsmin", "Source min level"),
                MetricDoc("Source Metrics", "mxsmax", "Source max levl"),
                MetricDoc("Source Metrics", "mxspkdb", "Source Pk lev dB"),
                MetricDoc("Source Metrics", "mxslrms", "Source RMS lev dB"),
                MetricDoc("Source Metrics", "mxspkr", "Source RMS Pk dB"),
                MetricDoc("Source Metrics", "mxstr", "Source RMS Tr dB"),
                MetricDoc("Source Metrics", "mxscr", "Source Crest factor"),
                MetricDoc("Source Metrics", "mxsfl", "Source Flat factor"),
                MetricDoc("Source Metrics", "mxspkc", "Source Pk count"),
                MetricDoc("Noise Metrics", "ndco", "Noise DC offset"),
                MetricDoc("Noise Metrics", "nmin", "Noise min level"),
                MetricDoc("Noise Metrics", "nmax", "Noise max levl"),
                MetricDoc("Noise Metrics", "npkdb", "Noise Pk lev dB"),
                MetricDoc("Noise Metrics", "nlrms", "Noise RMS lev dB"),
                MetricDoc("Noise Metrics", "npkr", "Noise RMS Pk dB"),
                MetricDoc("Noise Metrics", "ntr", "Noise RMS Tr dB"),
                MetricDoc("Noise Metrics", "ncr", "Noise Crest factor"),
                MetricDoc("Noise Metrics", "nfl", "Noise Flat factor"),
                MetricDoc("Noise Metrics", "npkc", "Noise Pk count"),
                MetricDoc(
                    "Truth Metrics",
                    "sedavg",
                    "(not implemented) Average SED activity over all frames [truth_parameters, 1]",
                ),
                MetricDoc(
                    "Truth Metrics",
                    "sedcnt",
                    "(not implemented) Count in number of frames that SED is active [truth_parameters, 1]",
                ),
                MetricDoc(
                    "Truth Metrics",
                    "sedtop3",
                    "(not implemented) 3 most active by largest sedavg [3, 1]",
                ),
                MetricDoc(
                    "Truth Metrics",
                    "sedtopn",
                    "(not implemented) N most active by largest sedavg [N, 1]",
                ),
            ]
        )
        for name in self.asr_configs:
            metrics.append(
                MetricDoc(
                    "Source Metrics",
                    f"mxsasr.{name}",
                    f"Mixture Source ASR text using {name} ASR as defined in mixdb asr_configs parameter",
                )
            )
            metrics.append(
                MetricDoc(
                    "Sources Metrics",
                    f"sasr.{name}",
                    f"Sources ASR text using {name} ASR as defined in mixdb asr_configs parameter",
                )
            )
            metrics.append(
                MetricDoc(
                    "Mixture Metrics",
                    f"mxasr.{name}",
                    f"ASR text using {name} ASR as defined in mixdb asr_configs parameter",
                )
            )
            metrics.append(
                MetricDoc(
                    "Sources Metrics",
                    f"basewer.{name}",
                    f"Word error rate of sasr.{name} vs. speech text metadata for the source",
                )
            )
            metrics.append(
                MetricDoc(
                    "Mixture Metrics",
                    f"mxwer.{name}",
                    f"Word error rate of mxasr.{name} vs. sasr.{name}",
                )
            )

        return metrics

    @cached_property
    def class_balancing(self) -> bool:
        with self.db() as c:
            return bool(c.execute("SELECT class_balancing FROM top").fetchone()[0])

    @cached_property
    def feature(self) -> str:
        with self.db() as c:
            return str(c.execute("SELECT feature FROM top").fetchone()[0])

    @cached_property
    def fg_decimation(self) -> int:
        return self.fg_info.decimation

    @cached_property
    def fg_stride(self) -> int:
        return self.fg_info.stride

    @cached_property
    def fg_step(self) -> int:
        return self.fg_info.step

    @cached_property
    def feature_parameters(self) -> int:
        return self.fg_info.feature_parameters

    @cached_property
    def ft_config(self) -> TransformConfig:
        return self.fg_info.ft_config

    @cached_property
    def eft_config(self) -> TransformConfig:
        return self.fg_info.eft_config

    @cached_property
    def it_config(self) -> TransformConfig:
        return self.fg_info.it_config

    @cached_property
    def transform_frame_ms(self) -> float:
        from ..constants import SAMPLE_RATE

        return float(self.ft_config.overlap) / float(SAMPLE_RATE / 1000)

    @cached_property
    def feature_ms(self) -> float:
        return self.transform_frame_ms * self.fg_decimation * self.fg_stride

    @cached_property
    def feature_samples(self) -> int:
        return self.ft_config.overlap * self.fg_decimation * self.fg_stride

    @cached_property
    def feature_step_ms(self) -> float:
        return self.transform_frame_ms * self.fg_decimation * self.fg_step

    @cached_property
    def feature_step_samples(self) -> int:
        return self.ft_config.overlap * self.fg_decimation * self.fg_step

    def total_samples(self, m_ids: GeneralizedIDs = "*") -> int:
        return sum([self.mixture(m_id).samples for m_id in self.mixids_to_list(m_ids)])

    def total_transform_frames(self, m_ids: GeneralizedIDs = "*") -> int:
        return self.total_samples(m_ids) // self.ft_config.overlap

    def total_feature_frames(self, m_ids: GeneralizedIDs = "*") -> int:
        return self.total_samples(m_ids) // self.feature_step_samples

    def mixture_transform_frames(self, m_id: int) -> int:
        from .helpers import frames_from_samples

        return frames_from_samples(self.mixture(m_id).samples, self.ft_config.overlap)

    def mixture_feature_frames(self, m_id: int) -> int:
        from .helpers import frames_from_samples

        return frames_from_samples(self.mixture(m_id).samples, self.feature_step_samples)

    def mixids_to_list(self, m_ids: GeneralizedIDs = "*") -> list[int]:
        """Resolve generalized mixture IDs to a list of integers

        :param m_ids: Generalized mixture IDs
        :return: List of mixture ID integers
        """
        from .helpers import generic_ids_to_list

        return generic_ids_to_list(self.num_mixtures, m_ids)

    @cached_property
    def class_labels(self) -> list[str]:
        """Get class labels from db

        :return: Class labels
        """
        with self.db() as c:
            return [str(item[0]) for item in c.execute("SELECT label FROM class_label ORDER BY id").fetchall()]

    @cached_property
    def class_weights_thresholds(self) -> list[float]:
        """Get class weights thresholds from db

        :return: Class weights thresholds
        """
        with self.db() as c:
            return [float(item[0]) for item in c.execute("SELECT threshold FROM class_weights_threshold").fetchall()]

    def category_truth_configs(self, category: str) -> dict[str, str]:
        return _category_truth_configs(self.db, category, self.use_cache)

    def source_truth_configs(self, s_id: int) -> TruthConfigs:
        return _source_truth_configs(self.db, s_id, self.use_cache)

    def mixture_truth_configs(self, m_id: int) -> TruthsConfigs:
        mixture = self.mixture(m_id)
        return {
            category: self.source_truth_configs(mixture.all_sources[category].file_id)
            for category in mixture.all_sources
        }

    @cached_property
    def random_snrs(self) -> list[float]:
        """Get random snrs from db

        :return: Random SNRs
        """
        with self.db() as c:
            return list(
                {float(item[0]) for item in c.execute("SELECT snr FROM source WHERE snr_random == 1").fetchall()}
            )

    @cached_property
    def snrs(self) -> list[float]:
        """Get snrs from db

        :return: SNRs
        """
        with self.db() as c:
            return list(
                {float(item[0]) for item in c.execute("SELECT snr FROM source WHERE snr_random == 0").fetchall()}
            )

    @cached_property
    def all_snrs(self) -> list[UniversalSNR]:
        return sorted(
            set(
                [UniversalSNR(is_random=False, value=snr) for snr in self.snrs]
                + [UniversalSNR(is_random=True, value=snr) for snr in self.random_snrs]
            )
        )

    @cached_property
    def spectral_masks(self) -> list[SpectralMask]:
        """Get spectral masks from db

        :return: Spectral masks
        """
        from .db_datatypes import SpectralMaskRecord

        with self.db() as c:
            spectral_masks = [
                SpectralMaskRecord(*result) for result in c.execute("SELECT * FROM spectral_mask").fetchall()
            ]
            return [
                SpectralMask(
                    f_max_width=spectral_mask.f_max_width,
                    f_num=spectral_mask.f_num,
                    t_max_width=spectral_mask.t_max_width,
                    t_num=spectral_mask.t_num,
                    t_max_percent=spectral_mask.t_max_percent,
                )
                for spectral_mask in spectral_masks
            ]

    def spectral_mask(self, sm_id: int) -> SpectralMask:
        """Get spectral mask with ID from db

        :param sm_id: Spectral mask ID
        :return: Spectral mask
        """
        return _spectral_mask(self.db, sm_id, self.use_cache)

    @cached_property
    def source_files(self) -> dict[str, list[SourceFile]]:
        """Get source files from db

        :return: Source files
        """
        import json

        from ..datatypes import TruthConfig
        from ..datatypes import TruthConfigs
        from .db_datatypes import SourceFileRecord

        with self.db() as c:
            source_files: dict[str, list[SourceFile]] = {}
            categories = c.execute("SELECT DISTINCT category FROM source_file").fetchall()
            for category in categories:
                source_files[category[0]] = []
                source_file_records = [
                    SourceFileRecord(*result)
                    for result in c.execute("SELECT * FROM source_file WHERE ? = category", (category[0],)).fetchall()
                ]
                for source_file_record in source_file_records:
                    truth_configs: TruthConfigs = {}
                    for truth_config_records in c.execute(
                        """
                        SELECT truth_config.config
                        FROM truth_config, source_file_truth_config
                        WHERE ? = source_file_truth_config.source_file_id
                        AND truth_config.id = source_file_truth_config.truth_config_id
                        """,
                        (source_file_record.id,),
                    ).fetchall():
                        truth_config = json.loads(truth_config_records[0])
                        truth_configs[truth_config["name"]] = TruthConfig(
                            function=truth_config["function"],
                            stride_reduction=truth_config["stride_reduction"],
                            config=truth_config["config"],
                        )
                    source_files[source_file_record.category].append(
                        SourceFile(
                            id=source_file_record.id,
                            category=source_file_record.category,
                            name=source_file_record.name,
                            samples=source_file_record.samples,
                            class_indices=json.loads(source_file_record.class_indices),
                            level_type=source_file_record.level_type,
                            truth_configs=truth_configs,
                            speaker_id=source_file_record.speaker_id,
                        )
                    )
            return source_files

    @cached_property
    def source_file_ids(self) -> dict[str, list[int]]:
        """Get source file IDs from db

        :return: Dictionary of list of source file IDs
        """
        with self.db() as c:
            source_file_ids: dict[str, list[int]] = {}
            categories = c.execute("SELECT DISTINCT category FROM source_file").fetchall()
            for category in categories:
                source_file_ids[category[0]] = [
                    int(item[0])
                    for item in c.execute("SELECT id FROM source_file WHERE ? = category", (category[0],)).fetchall()
                ]
            return source_file_ids

    def source_file(self, s_id: int) -> SourceFile:
        """Get source file with ID from db

        :param s_id: Source file ID
        :return: Source file
        """
        return _source_file(self.db, s_id, self.use_cache)

    def num_source_files(self, category: str) -> int:
        """Get number of source files from category from db

        :param category: Source category
        :return: Number of source files
        """
        return _num_source_files(self.db, category, self.use_cache)

    @cached_property
    def ir_files(self) -> list[ImpulseResponseFile]:
        """Get impulse response files from db

        :return: Impulse response files
        """
        from .db_datatypes import ImpulseResponseFileRecord

        with self.db() as c:
            files: list[ImpulseResponseFile] = []
            entries = c.execute("SELECT * FROM ir_file").fetchall()
            for entry in entries:
                file = ImpulseResponseFileRecord(*entry)

                tags = [
                    tag[0]
                    for tag in c.execute(
                        """
                SELECT ir_tag.tag
                FROM ir_tag, ir_file_ir_tag
                WHERE ? = ir_file_ir_tag.file_id
                AND ir_tag.id = ir_file_ir_tag.tag_id
                """,
                        (file.id,),
                    ).fetchall()
                ]

                files.append(
                    ImpulseResponseFile(
                        delay=file.delay,
                        name=file.name,
                        tags=tags,
                    )
                )

            return files

    @cached_property
    def ir_file_ids(self) -> list[int]:
        """Get impulse response file IDs from db

        :return: List of impulse response file IDs
        """
        with self.db() as c:
            return [int(item[0]) for item in c.execute("SELECT id FROM ir_file").fetchall()]

    def ir_file_ids_for_tag(self, tag: str) -> list[int]:
        """Get impulse response file IDs for given tag from db

        :return: List of impulse response file IDs for given tag
        """
        with self.db() as c:
            tag_id = c.execute("SELECT id FROM ir_tag WHERE ? = tag", (tag,)).fetchone()
            if not tag_id:
                return []

            return [
                int(item[0] - 1)
                for item in c.execute("SELECT file_id FROM ir_file_ir_tag WHERE ? = tag_id", (tag_id[0],)).fetchall()
            ]

    def ir_file(self, ir_id: int) -> str:
        """Get impulse response file name with ID from db

        :param ir_id: Impulse response file ID
        :return: Impulse response file name
        """
        return _ir_file(self.db, ir_id, self.use_cache)

    def ir_delay(self, ir_id: int) -> int:
        """Get impulse response delay with ID from db

        :param ir_id: Impulse response file ID
        :return: Impulse response delay
        """
        return _ir_delay(self.db, ir_id, self.use_cache)

    @cached_property
    def num_ir_files(self) -> int:
        """Get number of impulse response files from db

        :return: Number of impulse response files
        """
        with self.db() as c:
            return int(c.execute("SELECT count(id) FROM ir_file").fetchone()[0])

    @cached_property
    def ir_tags(self) -> list[str]:
        """Get tags of impulse response files from db

        :return: Tags of impulse response files
        """
        with self.db() as c:
            return [tag[0] for tag in c.execute("SELECT tag FROM ir_tag").fetchall()]

    @property
    def mixtures(self) -> list[Mixture]:
        """Get mixtures from db

        :return: Mixtures
        """
        from .db_datatypes import MixtureRecord
        from .db_datatypes import SourceRecord
        from .helpers import to_mixture
        from .helpers import to_source

        with self.db() as c:
            mixtures: list[Mixture] = []
            for mixture in [MixtureRecord(*record) for record in c.execute("SELECT * FROM mixture").fetchall()]:
                sources_list = [
                    to_source(SourceRecord(*source))
                    for source in c.execute(
                        """
                        SELECT source.*
                        FROM source, mixture_source
                        WHERE ? = mixture_source.mixture_id AND source.id = mixture_source.source_id
                        """,
                        (mixture.id,),
                    ).fetchall()
                ]

                sources: Sources = {}
                for source in sources_list:
                    sources[self.source_file(source.file_id).category] = source

                mixtures.append(to_mixture(mixture, sources))

            return mixtures

    @cached_property
    def mixture_ids(self) -> list[int]:
        """Get mixture IDs from db

        :return: List of zero-based mixture IDs
        """
        with self.db() as c:
            return [int(item[0]) - 1 for item in c.execute("SELECT id FROM mixture").fetchall()]

    def mixture(self, m_id: int) -> Mixture:
        """Get mixture record with ID from db

        :param m_id: Zero-based mixture ID
        :return: Mixture record
        """
        return _mixture(self.db, m_id, self.use_cache)

    @cached_property
    def mixid_width(self) -> int:
        with self.db() as c:
            return int(c.execute("SELECT mixid_width FROM top").fetchone()[0])

    def mixture_location(self, m_id: int) -> str:
        """Get the file location for the give mixture ID

        :param m_id: Zero-based mixture ID
        :return: File location
        """
        from os.path import join

        return join(self.location, self.mixture(m_id).name)

    @cached_property
    def num_mixtures(self) -> int:
        """Get number of mixtures from db

        :return: Number of mixtures
        """
        with self.db() as c:
            return int(c.execute("SELECT count(id) FROM mixture").fetchone()[0])

    def read_mixture_data(self, m_id: int, items: list[str] | str) -> dict[str, Any]:
        """Read mixture data

        :param m_id: Zero-based mixture ID
        :param items: String(s) of dataset(s) to retrieve
        :return: Dictionary of name: data
        """
        from .data_io import read_cached_data

        return read_cached_data(self.location, "mixture", self.mixture(m_id).name, items)

    def read_source_audio(self, s_id: int) -> AudioT:
        """Read source audio

        :param s_id: Source ID
        :return: Source audio
        """
        from .audio import read_audio

        return read_audio(self.source_file(s_id).name, self.use_cache)

    def mixture_class_indices(self, m_id: int) -> list[int]:
        class_indices: list[int] = []
        for s_id in self.mixture(m_id).source_ids.values():
            class_indices.extend(self.source_file(s_id).class_indices)
        return sorted(set(class_indices))

    def mixture_sources(self, m_id: int, force: bool = False, cache: bool = False) -> SourcesAudioT:
        """Get the pre-truth source audio data (one per source in the mixture) for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param force: Force computing data from original sources regardless of whether cached data exists
        :param cache: Cache result
        :return: Dictionary of pre-truth source audio data (one per source in the mixture)
        """
        from .data_io import write_cached_data
        from .effects import apply_effects
        from .effects import conform_audio_to_length

        if not force:
            sources = self.read_mixture_data(m_id, "sources")["sources"]
            if sources is not None:
                return sources

        mixture = self.mixture(m_id)
        if mixture is None:
            raise ValueError(f"Could not find mixture for m_id: {m_id}")

        sources = {}
        for category, source in mixture.all_sources.items():
            source = mixture.all_sources[category]
            audio = self.read_source_audio(source.file_id)
            audio = apply_effects(self, audio, source.effects, pre=True, post=False)
            audio = conform_audio_to_length(audio, mixture.samples, source.repeat, source.start)
            sources[category] = audio

        if cache:
            write_cached_data(
                location=self.location,
                name="mixture",
                index=mixture.name,
                items={"sources": sources},
            )

        return sources

    def mixture_sources_f(
        self,
        m_id: int,
        sources: SourcesAudioT | None = None,
        force: bool = False,
        cache: bool = False,
    ) -> SourcesAudioF:
        """Get the pre-truth source transform data (one per source in the mixture) for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param sources: Dictionary of pre-truth source audio data (one per source in the mixture)
        :param force: Force computing data from original sources regardless of whether cached data exists
        :param cache: Cache result
        :return: Dictionary of pre-truth source transform data (one per source in the mixture)
        """
        from .data_io import write_cached_data
        from .helpers import forward_transform

        if sources is None:
            sources = self.mixture_sources(m_id, force)

        sources_f = {category: forward_transform(sources[category], self.ft_config) for category in sources}

        if cache:
            write_cached_data(
                location=self.location,
                name="mixture",
                index=self.mixture(m_id).name,
                items={"sources_f": sources_f},
            )

        return sources_f

    def mixture_source(
        self,
        m_id: int,
        sources: SourcesAudioT | None = None,
        force: bool = False,
        cache: bool = False,
    ) -> AudioT:
        """Get the post-truth, summed, and gained source audio data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param sources: Dictionary of pre-truth source audio data (one per source in the mixture)
        :param force: Force computing data from original sources regardless of whether cached data exists
        :param cache: Cache result
        :return: Post-truth, gained, and summed source audio data
        """
        import numpy as np

        from .data_io import write_cached_data
        from .effects import apply_effects

        if not force:
            source = self.read_mixture_data(m_id, "source")["source"]
            if source is not None:
                return source

        if sources is None:
            sources = self.mixture_sources(m_id, force)

        mixture = self.mixture(m_id)

        source = np.sum(
            [
                apply_effects(
                    self,
                    audio=sources[category],
                    effects=mixture.all_sources[category].effects,
                    pre=False,
                    post=True,
                )
                * mixture.all_sources[category].snr_gain
                for category in sources
                if category != "noise"
            ],
            axis=0,
        )

        if cache:
            write_cached_data(
                location=self.location,
                name="mixture",
                index=mixture.name,
                items={"source": source},
            )

        return source

    def mixture_source_f(
        self,
        m_id: int,
        sources: SourcesAudioT | None = None,
        source: AudioT | None = None,
        force: bool = False,
        cache: bool = False,
    ) -> AudioF:
        """Get the post-truth, summed, and gained source transform data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param sources: Dictionary of pre-truth source audio data (one per source in the mixture)
        :param source: Post-truth, gained, and summed source audio for the given m_id
        :param force: Force computing data from original sources regardless of whether cached data exists
        :param cache: Cache result
        :return: Post-truth, gained, and summed source transform data
        """
        from .data_io import write_cached_data
        from .helpers import forward_transform

        if source is None:
            source = self.mixture_source(m_id, sources, force)

        source_f = forward_transform(source, self.ft_config)

        if cache:
            write_cached_data(
                location=self.location,
                name="mixture",
                index=self.mixture(m_id).name,
                items={"source_f": source_f},
            )

        return source_f

    def mixture_noise(
        self,
        m_id: int,
        sources: SourcesAudioT | None = None,
        force: bool = False,
        cache: bool = False,
    ) -> AudioT:
        """Get the post-truth and gained noise audio data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param sources: Dictionary of pre-truth source audio data (one per source in the mixture)
        :param force: Force computing data from original sources regardless of whether cached data exists
        :param cache: Cache result
        :return: Post-truth and gained noise audio data
        """
        from .data_io import write_cached_data
        from .effects import apply_effects

        if not force:
            noise = self.read_mixture_data(m_id, "noise")["noise"]
            if noise is not None:
                return noise

        if sources is None:
            sources = self.mixture_sources(m_id, force)

        noise = self.mixture(m_id).noise
        noise = apply_effects(self, sources["noise"], noise.effects, pre=False, post=True) * noise.snr_gain

        if cache:
            write_cached_data(
                location=self.location,
                name="mixture",
                index=self.mixture(m_id).name,
                items={"noise": noise},
            )

        return noise

    def mixture_noise_f(
        self,
        m_id: int,
        sources: SourcesAudioT | None = None,
        noise: AudioT | None = None,
        force: bool = False,
        cache: bool = False,
    ) -> AudioF:
        """Get the post-truth and gained noise transform for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param sources: Dictionary of pre-truth source audio data (one per source in the mixture)
        :param noise: Post-truth and gained noise audio data
        :param force: Force computing data from original sources regardless of whether cached data exists
        :param cache: Cache result
        :return: Post-truth and gained noise transform data
        """
        from .data_io import write_cached_data
        from .helpers import forward_transform

        if force or noise is None:
            noise = self.mixture_noise(m_id, sources, force)

        noise_f = forward_transform(noise, self.ft_config)
        if cache:
            write_cached_data(
                location=self.location,
                name="mixture",
                index=self.mixture(m_id).name,
                items={"noise_f": noise_f},
            )

        return noise_f

    def mixture_mixture(
        self,
        m_id: int,
        sources: SourcesAudioT | None = None,
        source: AudioT | None = None,
        noise: AudioT | None = None,
        force: bool = False,
        cache: bool = False,
    ) -> AudioT:
        """Get the mixture audio data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param sources: Dictionary of pre-truth source audio data (one per source in the mixture)
        :param source: Post-truth, gained, and summed source audio data
        :param noise: Post-truth and gained noise audio data
        :param force: Force computing data from original sources regardless of whether cached data exists
        :param cache: Cache result
        :return: Mixture audio data
        """
        from .data_io import write_cached_data

        if not force:
            mixture = self.read_mixture_data(m_id, "mixture")["mixture"]
            if mixture is not None:
                return mixture

        if source is None:
            source = self.mixture_source(m_id, sources, force)

        if noise is None:
            noise = self.mixture_noise(m_id, sources, force)

        mixture = source + noise

        if cache:
            write_cached_data(
                location=self.location,
                name="mixture",
                index=self.mixture(m_id).name,
                items={"mixture": mixture},
            )

        return mixture

    def mixture_mixture_f(
        self,
        m_id: int,
        sources: SourcesAudioT | None = None,
        source: AudioT | None = None,
        noise: AudioT | None = None,
        mixture: AudioT | None = None,
        force: bool = False,
        cache: bool = False,
    ) -> AudioF:
        """Get the mixture transform for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param sources: Dictionary of pre-truth source audio data (one per source in the mixture)
        :param source: Post-truth, gained, and summed source audio data
        :param noise: Post-truth and gained noise audio data
        :param mixture: Mixture audio data
        :param force: Force computing data from original sources regardless of whether cached data exists
        :param cache: Cache result
        :return: Mixture transform data
        """
        from .data_io import write_cached_data
        from .helpers import forward_transform
        from .spectral_mask import apply_spectral_mask

        if mixture is None:
            mixture = self.mixture_mixture(m_id, sources, source, noise, force)

        mixture_f = forward_transform(mixture, self.ft_config)

        m = self.mixture(m_id)
        if m.spectral_mask_id is not None:
            mixture_f = apply_spectral_mask(
                audio_f=mixture_f,
                spectral_mask=self.spectral_mask(int(m.spectral_mask_id)),
                seed=m.spectral_mask_seed,
            )

        if cache:
            write_cached_data(
                location=self.location,
                name="mixture",
                index=self.mixture(m_id).name,
                items={"mixture_f": mixture_f},
            )

        return mixture_f

    def mixture_truth_t(self, m_id: int, force: bool = False, cache: bool = False) -> TruthsDict:
        """Get the truth_t data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param force: Force computing data from original sources regardless of whether cached data exists
        :param cache: Cache result
        :return: list of truth_t data
        """
        from .data_io import write_cached_data
        from .truth import truth_function

        if not force:
            truth_t = self.read_mixture_data(m_id, "truth_t")["truth_t"]
            if truth_t is not None:
                return truth_t

        truth_t = truth_function(self, m_id)

        if cache:
            write_cached_data(
                location=self.location,
                name="mixture",
                index=self.mixture(m_id).name,
                items={"truth_t": truth_t},
            )

        return truth_t

    def mixture_segsnr_t(
        self,
        m_id: int,
        sources: SourcesAudioT | None = None,
        source: AudioT | None = None,
        noise: AudioT | None = None,
        force: bool = False,
        cache: bool = False,
    ) -> Segsnr:
        """Get the segsnr_t data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param sources: Dictionary of pre-truth source audio data (one per source in the mixture)
        :param source: Post-truth, gained, and summed source audio data
        :param noise: Post-truth and gained noise audio data
        :param force: Force computing data from original sources regardless of whether cached data exists
        :param cache: Cache result
        :return: segsnr_t data
        """
        import numpy as np
        import torch
        from pyaaware import ForwardTransform

        from .data_io import write_cached_data

        if not force:
            segsnr_t = self.read_mixture_data(m_id, "segsnr_t")["segsnr_t"]
            if segsnr_t is not None:
                return segsnr_t

        if source is None:
            source = self.mixture_source(m_id, sources, force)

        if noise is None:
            noise = self.mixture_noise(m_id, sources, force)

        ft = ForwardTransform(
            length=self.ft_config.length,
            overlap=self.ft_config.overlap,
            bin_start=self.ft_config.bin_start,
            bin_end=self.ft_config.bin_end,
            ttype=self.ft_config.ttype,
        )

        mixture = self.mixture(m_id)

        segsnr_t = np.empty(mixture.samples, dtype=np.float32)

        source_energy = ft.execute_all(torch.from_numpy(source))[1].numpy()
        noise_energy = ft.execute_all(torch.from_numpy(noise))[1].numpy()

        offsets = range(0, mixture.samples, self.ft_config.overlap)
        if len(source_energy) != len(offsets):
            raise ValueError(
                f"Number of frames in energy, {len(source_energy)}, is not number of frames in mixture, {len(offsets)}"
            )

        for idx, offset in enumerate(offsets):
            indices = slice(offset, offset + self.ft_config.overlap)

            if noise_energy[idx] == 0:
                snr = np.float32(np.inf)
            else:
                snr = np.float32(source_energy[idx] / noise_energy[idx])

            segsnr_t[indices] = snr

        if cache:
            write_cached_data(
                location=self.location,
                name="mixture",
                index=mixture.name,
                items={"segsnr_t": segsnr_t},
            )

        return segsnr_t

    def mixture_segsnr(
        self,
        m_id: int,
        segsnr_t: Segsnr | None = None,
        sources: SourcesAudioT | None = None,
        source: AudioT | None = None,
        noise: AudioT | None = None,
        force: bool = False,
        cache: bool = False,
    ) -> Segsnr:
        """Get the segsnr data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param segsnr_t: segsnr_t data
        :param sources: Dictionary of pre-truth source audio data (one per source in the mixture)
        :param source: Post-truth, gained, and summed source audio data
        :param noise: Post-truth and gained noise audio data
        :param force: Force computing data from original sources regardless of whether cached data exists
        :param cache: Cache result
        :return: segsnr data
        """
        from .data_io import write_cached_data

        if not force:
            segsnr = self.read_mixture_data(m_id, "segsnr")["segsnr"]
            if segsnr is not None:
                return segsnr

        if segsnr_t is None:
            segsnr_t = self.mixture_segsnr_t(m_id, sources, source, noise, force)

        segsnr = segsnr_t[0 :: self.ft_config.overlap]

        if cache:
            write_cached_data(
                location=self.location,
                name="mixture",
                index=self.mixture(m_id).name,
                items={"segsnr": segsnr},
            )

        return segsnr

    def mixture_ft(
        self,
        m_id: int,
        sources: SourcesAudioT | None = None,
        source: AudioT | None = None,
        noise: AudioT | None = None,
        mixture_f: AudioF | None = None,
        mixture: AudioT | None = None,
        truth_t: TruthsDict | None = None,
        force: bool = False,
        cache: bool = False,
    ) -> tuple[Feature, TruthsDict]:
        """Get the feature and truth_f data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param sources: Dictionary of pre-truth source audio data (one per source in the mixture)
        :param source: Post-truth, gained, and summed source audio data
        :param noise: Post-truth and gained noise audio data
        :param mixture_f: Mixture transform data
        :param mixture: Mixture audio data
        :param truth_t: truth_t
        :param force: Force computing data from original sources regardless of whether cached data exists
        :param cache: Cache result
        :return: Tuple of (feature, truth_f) data
        """
        from pyaaware import FeatureGenerator

        from .data_io import write_cached_data
        from .truth import truth_stride_reduction

        if not force:
            ft = self.read_mixture_data(m_id, ["feature", "truth_f"])
            if ft["feature"] is not None and ft["truth_f"] is not None:
                return ft["feature"], ft["truth_f"]

        if mixture_f is None:
            mixture_f = self.mixture_mixture_f(
                m_id=m_id,
                sources=sources,
                source=source,
                noise=noise,
                mixture=mixture,
                force=force,
            )

        if truth_t is None:
            truth_t = self.mixture_truth_t(m_id, force)

        fg = FeatureGenerator(self.fg_config.feature_mode, self.fg_config.truth_parameters)

        feature, truth_f = fg.execute_all(mixture_f, truth_t)
        if truth_f is None:
            raise TypeError("Unexpected truth of None from feature generator")

        truth_configs = self.mixture_truth_configs(m_id)
        for category, configs in truth_configs.items():
            for name, config in configs.items():
                if self.truth_parameters[category][name] is not None:
                    truth_f[category][name] = truth_stride_reduction(truth_f[category][name], config.stride_reduction)

        if cache:
            write_cached_data(
                location=self.location,
                name="mixture",
                index=self.mixture(m_id).name,
                items={"feature": truth_f, "truth_f": truth_f},
            )

        return feature, truth_f

    def mixture_feature(
        self,
        m_id: int,
        sources: SourcesAudioT | None = None,
        noise: AudioT | None = None,
        mixture: AudioT | None = None,
        truth_t: TruthsDict | None = None,
        force: bool = False,
        cache: bool = False,
    ) -> Feature:
        """Get the feature data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param sources: Dictionary of pre-truth source audio data (one per source in the mixture)
        :param noise: Post-truth and gained noise audio data
        :param mixture: Mixture audio data
        :param truth_t: truth_t
        :param force: Force computing data from original sources regardless of whether cached data exists
        :param cache: Cache result
        :return: Feature data
        """
        from .data_io import write_cached_data

        feature = self.mixture_ft(
            m_id=m_id,
            sources=sources,
            noise=noise,
            mixture=mixture,
            truth_t=truth_t,
            force=force,
        )[0]

        if cache:
            write_cached_data(
                location=self.location,
                name="mixture",
                index=self.mixture(m_id).name,
                items={"feature": feature},
            )

        return feature

    def mixture_truth_f(
        self,
        m_id: int,
        sources: SourcesAudioT | None = None,
        noise: AudioT | None = None,
        mixture: AudioT | None = None,
        truth_t: TruthsDict | None = None,
        force: bool = False,
        cache: bool = False,
    ) -> TruthDict:
        """Get the truth_f data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param sources: Dictionary of pre-truth source audio data (one per source in the mixture)
        :param noise: Post-truth and gained noise audio data
        :param mixture: Mixture audio data
        :param truth_t: truth_t
        :param force: Force computing data from original sources regardless of whether cached data exists
        :param cache: Cache result
        :return: truth_f data
        """
        from .data_io import write_cached_data

        truth_f = self.mixture_ft(
            m_id=m_id,
            sources=sources,
            noise=noise,
            mixture=mixture,
            truth_t=truth_t,
            force=force,
        )[1]

        if cache:
            write_cached_data(
                location=self.location,
                name="mixture",
                index=self.mixture(m_id).name,
                items={"truth_f": truth_f},
            )

        return truth_f

    def mixture_class_count(self, m_id: int, truth_t: TruthsDict | None = None) -> dict[str, ClassCount]:
        """Compute the number of frames for which each class index is active for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param truth_t: truth_t
        :return: Dictionary of class counts
        """
        import numpy as np

        if truth_t is None:
            truth_t = self.mixture_truth_t(m_id)

        class_count: dict[str, ClassCount] = {}

        truth_configs = self.mixture_truth_configs(m_id)
        for category in truth_configs:
            class_count[category] = [0] * self.num_classes
            for configs in truth_configs[category]:
                if "sed" in configs:
                    for cl in range(self.num_classes):
                        class_count[category][cl] = int(
                            np.sum(truth_t[category]["sed"][:, cl] >= self.class_weights_thresholds[cl])
                        )

        return class_count

    @cached_property
    def speaker_metadata_tiers(self) -> list[str]:
        import json

        with self.db() as c:
            return json.loads(c.execute("SELECT speaker_metadata_tiers FROM top WHERE 1 = id").fetchone()[0])

    @cached_property
    def textgrid_metadata_tiers(self) -> list[str]:
        import json

        with self.db() as c:
            return json.loads(c.execute("SELECT textgrid_metadata_tiers FROM top WHERE 1 = id").fetchone()[0])

    @cached_property
    def speech_metadata_tiers(self) -> list[str]:
        return sorted(set(self.speaker_metadata_tiers + self.textgrid_metadata_tiers))

    def speaker(self, s_id: int | None, tier: str) -> str | None:
        return _speaker(self.db, s_id, tier, self.use_cache)

    def speech_metadata(self, tier: str) -> list[str]:
        from .helpers import get_textgrid_tier_from_source_file

        results: set[str] = set()
        if tier in self.textgrid_metadata_tiers:
            for source_files in self.source_files.values():
                for source_file in source_files:
                    data = get_textgrid_tier_from_source_file(source_file.name, tier)
                    if data is None:
                        continue
                    if isinstance(data, list):
                        for item in data:
                            results.add(item.label)
                    else:
                        results.add(data)
        elif tier in self.speaker_metadata_tiers:
            for source_files in self.source_files.values():
                for source_file in source_files:
                    data = self.speaker(source_file.speaker_id, tier)
                    if data is not None:
                        results.add(data)

        return sorted(results)

    def mixture_speech_metadata(self, mixid: int, tier: str) -> dict[str, SpeechMetadata]:
        from praatio.utilities.constants import Interval

        from .helpers import get_textgrid_tier_from_source_file

        results: dict[str, SpeechMetadata] = {}
        is_textgrid = tier in self.textgrid_metadata_tiers
        if is_textgrid:
            for category, source in self.mixture(mixid).all_sources.items():
                data = get_textgrid_tier_from_source_file(self.source_file(source.file_id).name, tier)
                if isinstance(data, list):
                    # Check for tempo effect and adjust Interval start and end data as needed
                    entries = []
                    for entry in data:
                        entries.append(
                            Interval(
                                entry.start / source.pre_tempo,
                                entry.end / source.pre_tempo,
                                entry.label,
                            )
                        )
                    results[category] = entries
                else:
                    results[category] = data
        else:
            for category, source in self.mixture(mixid).all_sources.items():
                results[category] = self.speaker(self.source_file(source.file_id).speaker_id, tier)

        return results

    def mixids_for_speech_metadata(
        self,
        tier: str | None = None,
        value: str | None = None,
        where: str | None = None,
    ) -> dict[str, list[int]]:
        """Get a list of mixture IDs for the given speech metadata tier.

        If 'where' is None, then include mixture IDs whose tier values are equal to the given 'value'.
        If 'where' is not None, then ignore 'value' and use the given SQL where clause to determine
        which entries to include.

        Examples:
        >>> mixdb = MixtureDatabase('/mixdb_location')

        >>> mixids = mixdb.mixids_for_speech_metadata('speaker_id', 'TIMIT_ABW0')
        Get mixture IDs for mixtures with speakers whose speaker_ids are 'TIMIT_ABW0'.

        >>> mixids = mixdb.mixids_for_speech_metadata(where='age >= 27')
        Get mixture IDs for mixtures with speakers whose ages are greater than or equal to 27.

        >>> mixids = mixdb.mixids_for_speech_metadata(where="dialect in ('New York City', 'Northern')")
        Get mixture IDs for mixtures with speakers whose dialects are either 'New York City' or 'Northern'.
        """
        if value is None and where is None:
            raise ValueError("Must provide either value or where")

        if where is None:
            if tier is None:
                raise ValueError("Must provide tier")
            where = f"{tier} = '{value}'"

        if tier is not None and tier in self.textgrid_metadata_tiers:
            raise ValueError(f"TextGrid tier data, '{tier}', is not supported in mixids_for_speech_metadata().")

        with self.db() as c:
            results = c.execute(f"SELECT id FROM speaker WHERE {where}").fetchall()
            speaker_ids = ",".join(map(str, [i[0] for i in results]))

            results = c.execute(f"SELECT id, category FROM source_file WHERE speaker_id IN ({speaker_ids})").fetchall()
            source_file_ids: dict[str, list[int]] = {}
            for result in results:
                source_file_id, category = result
                if category not in source_file_ids:
                    source_file_ids[category] = [source_file_id]
                else:
                    source_file_ids[category].append(source_file_id)

            mixids: dict[str, list[int]] = {}
            for category in source_file_ids:
                id_str = ",".join(map(str, source_file_ids[category]))
                results = c.execute(f"SELECT id FROM source WHERE file_id IN ({id_str})").fetchall()
                source_ids = ",".join(map(str, [i[0] for i in results]))

                results = c.execute(
                    f"SELECT mixture_id FROM mixture_source WHERE source_id IN ({source_ids})"
                ).fetchall()
                mixids[category] = [mixture_id[0] - 1 for mixture_id in results]

        return mixids

    def mixture_all_speech_metadata(self, m_id: int) -> dict[str, dict[str, SpeechMetadata]]:
        from .helpers import mixture_all_speech_metadata

        return mixture_all_speech_metadata(self, self.mixture(m_id))

    def cached_metrics(self, m_ids: GeneralizedIDs = "*") -> list[str]:
        """Get list of cached metrics for all mixtures."""
        from glob import glob
        from os.path import join
        from pathlib import Path

        supported_metrics = self.supported_metrics.names
        first = True
        result: set[str] = set()
        for m_id in self.mixids_to_list(m_ids):
            mixture_dir = join(self.location, "mixture", self.mixture(m_id).name)
            found = {Path(f).stem for f in glob(join(mixture_dir, "*.pkl"))}
            if first:
                first = False
                for f in found:
                    if f in supported_metrics:
                        result.add(f)
            else:
                result = result & found

        return sorted(result)

    def mixture_metrics(self, m_id: int, metrics: list[str], force: bool = False) -> dict[str, Any]:
        """Get metrics data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param metrics: List of metrics to get
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Dictionary of metric data
        """
        from ..metrics import calculate_metrics

        return calculate_metrics(self, m_id, metrics, force)


def _spectral_mask(db: partial, sm_id: int, use_cache: bool = True) -> SpectralMask:
    """Get spectral mask with ID from db

    :param db: Database context
    :param sm_id: Spectral mask ID
    :param use_cache: If true, use LRU caching
    :return: Spectral mask
    """
    if use_cache:
        return __spectral_mask(db, sm_id)
    return __spectral_mask.__wrapped__(db, sm_id)


@lru_cache
def __spectral_mask(db: partial, sm_id: int) -> SpectralMask:
    from .db_datatypes import SpectralMaskRecord

    with db() as c:
        spectral_mask = SpectralMaskRecord(*c.execute("SELECT * FROM spectral_mask WHERE ? = id", (sm_id,)).fetchone())
        return SpectralMask(
            f_max_width=spectral_mask.f_max_width,
            f_num=spectral_mask.f_num,
            t_max_width=spectral_mask.t_max_width,
            t_num=spectral_mask.t_num,
            t_max_percent=spectral_mask.t_max_percent,
        )


def _num_source_files(db: partial, category: str, use_cache: bool = True) -> int:
    """Get number of source files from category from db

    :param db: Database context
    :param category: Source category
    :param use_cache: If true, use LRU caching
    :return: Number of source files
    """
    if use_cache:
        return __num_source_files(db, category)
    return __num_source_files.__wrapped__(db, category)


@lru_cache
def __num_source_files(db: partial, category: str) -> int:
    """Get number of source files from category from db

    :param db: Database context
    :param category: Source category
    :return: Number of source files
    """
    with db() as c:
        return int(c.execute("SELECT count(id) FROM source_file WHERE ? = category", (category,)).fetchone()[0])


def _source_file(db: partial, s_id: int, use_cache: bool = True) -> SourceFile:
    """Get source file with ID from db

    :param db: Database context
    :param s_id: Source file ID
    :param use_cache: If true, use LRU caching
    :return: Source file
    """
    if use_cache:
        return __source_file(db, s_id, use_cache)
    return __source_file.__wrapped__(db, s_id, use_cache)


@lru_cache
def __source_file(db: partial, s_id: int, use_cache: bool = True) -> SourceFile:
    """Get source file with ID from db

    :param db: Database context
    :param s_id: Source file ID
    :param use_cache: If true, use LRU caching
    :return: Source file
    """
    import json

    from .db_datatypes import SourceFileRecord

    with db() as c:
        source_file = SourceFileRecord(*c.execute("SELECT * FROM source_file WHERE ? = id", (s_id,)).fetchone())

        return SourceFile(
            category=source_file.category,
            name=source_file.name,
            samples=source_file.samples,
            class_indices=json.loads(source_file.class_indices),
            level_type=source_file.level_type,
            truth_configs=_source_truth_configs(db, s_id, use_cache),
            speaker_id=source_file.speaker_id,
        )


def _ir_file(db: partial, ir_id: int, use_cache: bool = True) -> str:
    """Get impulse response file name with ID from db

    :param db: Database context
    :param ir_id: Impulse response file ID
    :param use_cache: If true, use LRU caching
    :return: Impulse response file name
    """
    if use_cache:
        return __ir_file(db, ir_id)
    return __ir_file.__wrapped__(db, ir_id)


@lru_cache
def __ir_file(db: partial, ir_id: int) -> str:
    with db() as c:
        return str(c.execute("SELECT name FROM ir_file WHERE ? = id ", (ir_id + 1,)).fetchone()[0])


def _ir_delay(db: partial, ir_id: int, use_cache: bool = True) -> int:
    """Get impulse response delay with ID from db

    :param db: Database context
    :param ir_id: Impulse response file ID
    :param use_cache: If true, use LRU caching
    :return: Impulse response delay
    """
    if use_cache:
        return __ir_delay(db, ir_id)
    return __ir_delay.__wrapped__(db, ir_id)


@lru_cache
def __ir_delay(db: partial, ir_id: int) -> int:
    with db() as c:
        return int(c.execute("SELECT delay FROM ir_file WHERE ? = id", (ir_id + 1,)).fetchone()[0])


def _mixture(db: partial, m_id: int, use_cache: bool = True) -> Mixture:
    """Get mixture record with ID from db

    :param db: Database context
    :param m_id: Zero-based mixture ID
    :param use_cache: If true, use LRU caching
    :return: Mixture record
    """
    if use_cache:
        return __mixture(db, m_id)
    return __mixture.__wrapped__(db, m_id)


@lru_cache
def __mixture(db: partial, m_id: int) -> Mixture:
    from .db_datatypes import MixtureRecord
    from .db_datatypes import SourceRecord
    from .helpers import to_mixture
    from .helpers import to_source

    with db() as c:
        mixture = MixtureRecord(*c.execute("SELECT * FROM mixture WHERE ? = id", (m_id + 1,)).fetchone())

        sources: Sources = {}
        for source in c.execute(
            """
                SELECT source.*
                FROM source, mixture_source
                WHERE ? = mixture_source.mixture_id AND source.id = mixture_source.source_id
                """,
            (mixture.id,),
        ).fetchall():
            s = SourceRecord(*source)
            category = c.execute("SELECT category FROM source_file WHERE ? = id", (s.file_id,)).fetchone()[0]
            sources[category] = to_source(s)

        return to_mixture(mixture, sources)


def _speaker(db: partial, s_id: int | None, tier: str, use_cache: bool = True) -> str | None:
    if use_cache:
        return __speaker(db, s_id, tier)
    return __speaker.__wrapped__(db, s_id, tier)


@lru_cache
def __speaker(db: partial, s_id: int | None, tier: str) -> str | None:
    if s_id is None:
        return None

    with db() as c:
        data = c.execute(f"SELECT {tier} FROM speaker WHERE ? = id", (s_id,)).fetchone()
        if data is None:
            return None
        if data[0] is None:
            return None
        return data[0]


def _category_truth_configs(db: partial, category: str, use_cache: bool = True) -> dict[str, str]:
    if use_cache:
        return __category_truth_configs(db, category)
    return __category_truth_configs.__wrapped__(db, category)


@lru_cache
def __category_truth_configs(db: partial, category: str) -> dict[str, str]:
    import json

    truth_configs: dict[str, str] = {}
    with db() as c:
        s_ids = c.execute("SELECT id FROM source_file WHERE ? = category", (category,)).fetchall()

        for s_id in s_ids:
            for truth_config_record in c.execute(
                """
                SELECT truth_config.config
                FROM truth_config, source_file_truth_config
                WHERE ? = source_file_truth_config.source_file_id AND truth_config.id = source_file_truth_config.truth_config_id
                """,
                (s_id[0],),
            ).fetchall():
                truth_config = json.loads(truth_config_record[0])
                truth_configs[truth_config["name"]] = truth_config["function"]
    return truth_configs


def _source_truth_configs(db: partial, s_id: int, use_cache: bool = True) -> TruthConfigs:
    if use_cache:
        return __source_truth_configs(db, s_id)
    return __source_truth_configs.__wrapped__(db, s_id)


@lru_cache
def __source_truth_configs(db: partial, s_id: int) -> TruthConfigs:
    import json

    from ..datatypes import TruthConfig

    truth_configs: TruthConfigs = {}
    with db() as c:
        for truth_config_record in c.execute(
            """
            SELECT truth_config.config
            FROM truth_config, source_file_truth_config
            WHERE ? = source_file_truth_config.source_file_id AND truth_config.id = source_file_truth_config.truth_config_id
            """,
            (s_id,),
        ).fetchall():
            truth_config = json.loads(truth_config_record[0])
            truth_configs[truth_config["name"]] = TruthConfig(
                function=truth_config["function"],
                stride_reduction=truth_config["stride_reduction"],
                config=truth_config["config"],
            )
    return truth_configs



from ..datatypes import AudioT


def asl_p56(audio: AudioT) -> float:
    """Implement ITU-T P.56 method B
    :param audio: audio for which to calculate active speech level
    :return: Active speech level mean square energy
    """
    import numpy as np
    import scipy.signal as signal

    from ..constants import SAMPLE_RATE

    eps = np.finfo(np.float32).eps

    # Time constant of smoothing in seconds
    T = 0.03

    # Coefficient of smoothing
    g = np.exp(-1 / (SAMPLE_RATE * T))

    # Hangover time in seconds
    H = 0.2
    # Rounded up to next integer
    H_samples = np.ceil(H * SAMPLE_RATE)

    # Margin in dB, difference between threshold and active speech level
    M = 15.9

    # Number of thresholds
    thresh_num = 15

    # Series of fixed threshold voltages to apply to the envelope. These are spaced
    # in geometric progression, at intervals of not more than 2:1 (6.02 dB), from a
    # value equal to about half the maximum code down to a value equal to one
    # quantizing interval or lower.
    c = 2 ** np.arange(-15, thresh_num - 15, dtype=np.float32)

    # Activity counter for each threshold
    a = np.full(thresh_num, -1)

    # Hangover counter for each threshold
    h = np.full(thresh_num, H_samples)

    # Long-term level square energy of audio
    sq = sum(np.square(audio))

    # Use a 2nd order IIR filter to detect the envelope q
    p = signal.lfilter([1 - g, 0], [1, -g], abs(audio))
    # q is the envelope, obtained from moving average of abs(audio) (with slight "hangover").
    q = signal.lfilter([1 - g, 0], [1, -g], p)

    for k in range(len(audio)):
        for j in range(thresh_num):
            if q[k] >= c[j]:
                a[j] = a[j] + 1
                h[j] = 0
            elif h[j] < H_samples:
                a[j] = a[j] + 1
                h[j] = h[j] + 1
            else:
                break
    asl_msq = 0
    if a[0] == -1:
        return asl_msq

    a += 2
    A_db1 = 10 * np.log10(sq / a[0] + eps)
    C_db1 = 20 * np.log10(c[0] + eps)
    if A_db1 - C_db1 < M:
        return asl_msq

    A_db = np.zeros(thresh_num)
    C_db = np.zeros(thresh_num)
    delta = np.zeros(thresh_num)
    A_db[0] = A_db1
    C_db[0] = C_db1
    delta[0] = A_db1 - C_db1

    for j in range(1, thresh_num):
        A_db[j] = 10 * np.log10(sq / (a[j] + eps) + eps)
        C_db[j] = 20 * np.log10(c[j] + eps)

    for j in range(1, thresh_num):
        if a[j] != 0:
            delta[j] = A_db[j] - C_db[j]
            if delta[j] <= M:
                # Interpolate to find the asl_ms_log
                asl_ms_log = _bin_interp(A_db[j], A_db[j - 1], C_db[j], C_db[j - 1], M, 0.5)
                # This is the mean square value NOT the RMS
                asl_msq = 10.0 ** (asl_ms_log / 10)
                break

    return asl_msq


def _bin_interp(u_cnt: float, l_cnt: float, u_thr: float, l_thr: float, margin: float, tol: float) -> float:
    tol = abs(tol)

    # Check if extreme counts are not already the true active value
    iter_num = 1
    if abs(u_cnt - u_thr - margin) < tol:
        return u_cnt

    if abs(l_cnt - l_thr - margin) < tol:
        return l_cnt

    # Initialize first middle for given (initial) bounds
    m_cnt = (u_cnt + l_cnt) / 2.0
    m_thr = (u_thr + l_thr) / 2.0

    while True:
        # Loop until diff falls inside the tolerance (-tol<=diff<=tol)
        diff = m_cnt - m_thr - margin
        if abs(diff) <= tol:
            break

        # If tolerance is not met up to 20 iterations, then relax the tolerance by 10
        iter_num += 1
        if iter_num > 20:
            tol = tol * 1.1

        if diff > tol:
            m_cnt = (u_cnt + m_cnt) / 2.0
            m_thr = (u_thr + m_thr) / 2.0
        elif diff < -tol:
            m_cnt = (m_cnt + l_cnt) / 2.0
            m_thr = (m_thr + l_thr) / 2.0

    return m_cnt



import numpy as np


def int16_to_float(x: np.ndarray) -> np.ndarray:
    """Convert an int16 array to a floating point array with range +/- 1"""
    return x.astype(np.float32) / 32768


def float_to_int16(x: np.ndarray) -> np.ndarray:
    """Convert a floating point array with range +/- 1 to an int16 array"""
    return (x * 32768).astype(np.int16)
